#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""The module provides support for On-The-Fly encoding for RTxxx devices."""

import logging
import os
from copy import deepcopy
from struct import pack
from typing import Any, Dict, List, Optional, Union

from crcmod.predefined import mkPredefinedCrcFun
from cryptography.hazmat.primitives.keywrap import aes_key_wrap

from spsdk import SPSDKError
from spsdk import version as spsdk_version
from spsdk.apps.utils.utils import filepath_from_config, get_key
from spsdk.exceptions import SPSDKValueError
from spsdk.utils.database import Database
from spsdk.utils.exceptions import SPSDKRegsErrorBitfieldNotFound
from spsdk.utils.images import BinaryImage
from spsdk.utils.misc import (
    align_block,
    load_binary,
    reverse_bits_in_bytes,
    split_data,
    value_to_bytes,
    value_to_int,
)
from spsdk.utils.registers import Registers
from spsdk.utils.schema_validator import ConfigTemplate, ValidationSchemas

from . import OTFAD_DATA_FOLDER, OTFAD_DATABASE_FILE, OTFAD_SCH_FILE, Counter, crypto_backend

logger = logging.getLogger(__name__)


class KeyBlob:
    """OTFAD KeyBlob: The class specifies AES key and counter initial value for specified address range.

    typedef struct KeyBlob
    {
        unsigned char key[kAesKeySizeBytes]; // 16 bytes, 128-bits, KEY[A15...A00]
        unsigned char ctr[kCtrSizeBytes];    // 8 bytes, 64-bits, CTR[C7...C0]
        unsigned int srtaddr;                // region start, SRTADDR[31 - 10]
        unsigned int endaddr;                // region end, ENDADDR[31 - 10]; lowest three bits are used as flags
        // end of 32-byte area covered by CRC
        unsigned int zero_fill;      // zeros
        unsigned int key_blob_crc32; // crc32 over 1st 32-bytes
        // end of 40 byte (5*64-bit) key blob data
        unsigned char expanded_wrap_data[8]; // 8 bytes, used for wrap expanded data
        // end of 48 byte (6*64-bit) wrap data
        unsigned char unused_filler[16]; // unused fill to 64 bytes
    } keyblob_t;
    """

    _START_ADDR_MASK = 0x400 - 1
    # Region addresses are modulo 1024
    # The address ends with RO, ADE, VLD bits. From this perspective, only
    # bits [9:3] must be set to 1. The rest is configurable.
    _END_ADDR_MASK = 0x3F8

    # Key flags mask: RO, ADE, VLD
    _KEY_FLAG_MASK = 0x07
    # This field signals that the entire set of context registers (CTXn_KEY[0-3], CTXn_CTR[0-1],
    # CTXn_RGD_W[0-1] are read-only and cannot be modified. This field is sticky and remains
    # asserted until the next system reset. SR[RRAM] provides another level of register access
    # control and is independent of the RO indicator.
    KEY_FLAG_READ_ONLY = 0x4
    # AES Decryption Enable: For accesses hitting in a valid context, this bit indicates if the fetched data is to be
    # decrypted or simply bypassed.
    KEY_FLAG_ADE = 0x2
    # Valid: This field signals if the context is valid or not.
    KEY_FLAG_VLD = 0x1

    # key length in bytes
    KEY_SIZE = 16
    # counter length in bytes
    CTR_SIZE = 8
    # len of counter init value for export
    _EXPORT_CTR_IV_SIZE = 8
    # this constant seems to be fixed for SB2.1
    _EXPORT_NBLOCKS_5 = 5
    # binary export size
    _EXPORT_KEY_BLOB_SIZE = 64
    # QSPI image alignment length, 512 is supposed to be the safe alignment level for any QSPI device
    # this means that all QSPI images generated by this tool will be sizes of multiple 512
    _IMAGE_ALIGNMENT = 512
    # Encryption block size
    _ENCRYPTION_BLOCK_SIZE = 16

    def __init__(
        self,
        start_addr: int,
        end_addr: int,
        key: Optional[bytes] = None,
        counter_iv: Optional[bytes] = None,
        key_flags: int = KEY_FLAG_VLD | KEY_FLAG_ADE,
        # for testing
        zero_fill: Optional[bytes] = None,
        crc: Optional[bytes] = None,
    ):
        """Constructor.

        :param start_addr: start address of the region
        :param end_addr: end address of the region
        :param key_flags: see KEY_FLAG_xxx constants; default flags: RO = 0, ADE = 1, VLD = 1
        :param key: optional AES key; None to use random value
        :param counter_iv: optional counter init value for AES; None to use random value
        :param binaries: optional data chunks of this key blob
        :param zero_fill: optional value for zero_fill (for testing only); None to use random value (recommended)
        :param crc: optional value for unused CRC fill (for testing only); None to use random value (recommended)
        :raises SPSDKError: Start or end address are not aligned
        :raises SPSDKError: When there is invalid key
        :raises SPSDKError: When there is invalid start/end address
        :raises SPSDKError: When key_flags exceeds mask
        """
        if key is None:
            key = crypto_backend().random_bytes(self.KEY_SIZE)
        if counter_iv is None:
            counter_iv = crypto_backend().random_bytes(self.CTR_SIZE)
        if (len(key) != self.KEY_SIZE) and (len(counter_iv) != self.CTR_SIZE):
            raise SPSDKError("Invalid key")
        if start_addr < 0 or start_addr > end_addr or end_addr > 0xFFFFFFFF:
            raise SPSDKError("Invalid start/end address")
        if key_flags & ~self._KEY_FLAG_MASK != 0:
            raise SPSDKError(f"key_flags exceeds mask {hex(self._KEY_FLAG_MASK)}")
        if (start_addr & self._START_ADDR_MASK) != 0:
            raise SPSDKError(
                f"Start address must be aligned to {hex(self._START_ADDR_MASK + 1)} boundary"
            )
        # if (end_addr & self._END_ADDR_MASK) != self._END_ADDR_MASK:
        #     raise SPSDKError(f"End address must be aligned to {hex(self._END_ADDR_MASK)} boundary")
        self.key = key
        self.ctr_init_vector = counter_iv
        self.start_addr = start_addr
        self.end_addr = end_addr
        self.key_flags = key_flags
        self.zero_fill = zero_fill
        self.crc_fill = crc

    def info(self) -> str:
        """Text info about the instance."""
        msg = ""
        msg += f"Key:        {self.key.hex()}\n"
        msg += f"Counter IV: {self.ctr_init_vector.hex()}\n"
        msg += f"Start Addr: {hex(self.start_addr)}\n"
        msg += f"End Addr:   {hex(self.end_addr)}\n"
        return msg

    def plain_data(self) -> bytes:
        """Plain data for selected key range.

        :return: key blob exported into binary form (serialization)
        :raises SPSDKError: Invalid value of zero fill parameter
        :raises SPSDKError: Invalid value crc
        :raises SPSDKError: Invalid length binary data
        """
        result = bytes()
        result += self.key
        result += self.ctr_init_vector
        result += pack("<I", self.start_addr)
        if self.end_addr or self.key_flags:
            end_addr_with_flags = (
                ((self.end_addr - 1) & ~self._KEY_FLAG_MASK) | self.key_flags | self._END_ADDR_MASK
            )
        else:
            end_addr_with_flags = 0
        result += pack("<I", end_addr_with_flags)
        header_crc: bytes = mkPredefinedCrcFun("crc-32-mpeg")(result).to_bytes(4, "little")
        # zero fill
        if self.zero_fill:
            if len(self.zero_fill) != 4:
                raise SPSDKError("Invalid value")
            result += self.zero_fill
        else:
            result += crypto_backend().random_bytes(4)
        # CRC is not used, use random value
        if self.crc_fill:
            if len(self.crc_fill) != 4:
                raise SPSDKError("Invalid value crc")
            result += self.crc_fill
        else:
            result += header_crc
        result += bytes([0] * 8)  # expanded_wrap_data
        result += bytes([0] * 16)  # unused filler
        if len(result) != 64:
            raise SPSDKError("Invalid length binary data")
        return result

    # pylint: disable=invalid-name
    def export(
        self,
        kek: Union[bytes, str],
        iv: bytes = bytes([0xA6] * 8),
        byte_swap_cnt: int = 0,
    ) -> bytes:
        """Creates key wrap for the key blob.

        :param kek: key to encode; 16 bytes long
        :param iv: counter initialization vector; 8 bytes; optional, OTFAD uses empty init value
        :param byte_swap_cnt: Encrypted keyblob reverse byte count, 0 means NO reversing is enabled
        :return: Serialized key blob
        :raises SPSDKError: If any parameter is not valid
        :raises SPSDKError: If length of kek is not valid
        :raises SPSDKError: If length of data is not valid
        """
        if isinstance(kek, str):
            kek = bytes.fromhex(kek)
        if len(kek) != 16:
            raise SPSDKError("Invalid length of kek")
        if len(iv) != self._EXPORT_CTR_IV_SIZE:
            raise SPSDKError("Invalid length of initialization vector")
        n = self._EXPORT_NBLOCKS_5
        plaintext = self.plain_data()  # input data to be encrypted
        if len(plaintext) < n * 8:
            raise SPSDKError("Invalid length of data to be encrypted")

        blobs = bytes()
        wrap = aes_key_wrap(kek, plaintext[:40])
        if byte_swap_cnt > 0:
            for i in range(0, len(wrap), byte_swap_cnt):
                blobs += wrap[i : i + byte_swap_cnt][::-1]
        else:
            blobs += wrap

        return align_block(
            blobs, self._EXPORT_KEY_BLOB_SIZE, padding=0
        )  # align to 64 bytes (0 padding)

    def _get_ctr_nonce(self) -> bytes:
        """Get the counter initial value for image encryption.

        :return: counter bytes
        :raises SPSDKError: If length of counter is not valid
        """
        #  CTRn_x[127-0] = {CTR_W0_x[C0...C3],    // 32 bits of pre-programmed CTR
        #  CTR_W1_x[C4...C7],                     // another 32 bits of CTR
        #  CTR_W0_x[C0...C3] ^ CTR_W1_x[C4...C7], // exclusive-OR of CTR values
        #  systemAddress[31-4], 0000b             // 0-modulo-16 system address */

        if len(self.ctr_init_vector) != 8:
            raise SPSDKError("Invalid length of counter init")

        result = bytearray(16)
        result[:4] = self.ctr_init_vector[:4]
        result[4:8] = self.ctr_init_vector[4:]
        for i in range(0, 4):
            result[8 + i] = self.ctr_init_vector[0 + i] ^ self.ctr_init_vector[4 + i]

        # result[15:12] = start_addr as a counter; nonce has these bytes zero and value passes as counter init value

        return bytes(result)

    def contains_addr(self, addr: int) -> bool:
        """Whether key blob contains specified address.

        :param addr: to be tested
        :return: True if yes, False otherwise
        """
        return self.start_addr <= addr <= self.end_addr

    def matches_range(self, image_start: int, image_end: int) -> bool:
        """Whether key blob matches address range of the image to be encrypted.

        :param image_start: start address of the image
        :param image_end: last address of the image
        :return: True if yes, False otherwise
        """
        return self.contains_addr(image_start) and self.contains_addr(image_end)

    def encrypt_image(
        self,
        base_address: int,
        data: bytes,
        byte_swap: bool,
        counter_value: Optional[int] = None,
    ) -> bytes:
        """Encrypt specified data.

        :param base_address: of the data in target memory; must be >= self.start_addr
        :param data: to be encrypted (e.g. plain image); base_address + len(data) must be <= self.end_addr
        :param byte_swap: this probably depends on the flash device, how bytes are organized there
                True should be used for FLASH on EVK RT6xx; False for FLASH on EVK RT5xx
        :param counter_value: Optional counter value, if not specified start address of keyblob will be used
        :return: encrypted data
        :raises SPSDKError: If start address is not valid
        """
        if base_address % 16 != 0:
            raise SPSDKError("Invalid start address")  # Start address has to be 16 byte aligned
        data = align_block(data, self._ENCRYPTION_BLOCK_SIZE)  # align data length
        data_len = len(data)

        # check start and end addresses
        # Support dual image boot, do not raise exception
        if not self.matches_range(base_address, base_address + data_len - 1):
            logger.warning(
                f"Image address range is not within key blob: "
                f"{hex(self.start_addr)}-{hex(self.end_addr)}."
                " Ignore this if flash remap feature is used"
            )
        result = bytes()
        # SPSDK-936 change base address to start address of keyblob to allow flash remap on RT5xx/RT6xx
        if not counter_value:
            counter_value = self.start_addr

        counter = Counter(
            self._get_ctr_nonce(), ctr_value=counter_value, ctr_byteorder_encoding="big"
        )

        for index in range(0, data_len, 16):
            # prepare data in byte order
            if byte_swap:
                # swap 8 bytes + swap 8 bytes
                data_2_encr = (
                    data[-data_len + index + 7 : -data_len + index - 1 : -1]
                    + data[-data_len + index + 15 : -data_len + index + 7 : -1]
                )
            else:
                data_2_encr = data[index : index + 16]
            # encrypt
            encr_data = crypto_backend().aes_ctr_encrypt(self.key, data_2_encr, counter.value)
            # fix byte order in result
            if byte_swap:
                result += encr_data[-9:-17:-1] + encr_data[-1:-9:-1]  # swap 8 bytes + swap 8 bytes
            else:
                result += encr_data
            # update counter for encryption
            counter.increment(16)

        if len(result) != data_len:
            raise SPSDKError("Invalid length of encrypted data")
        return bytes(result)

    @property
    def is_encrypted(self) -> bool:
        """Get the required encryption or not.

        :return: True if blob is encrypted, False otherwise.
        """
        return (bool)(
            (self.key_flags & (self.KEY_FLAG_ADE | self.KEY_FLAG_VLD))
            == (self.KEY_FLAG_ADE | self.KEY_FLAG_VLD)
        )


class Otfad:
    """OTFAD: On-the-Fly AES Decryption Module."""

    OTFAD_DATA_UNIT = 0x400

    def __init__(self) -> None:
        """Constructor."""
        self._key_blobs: List[KeyBlob] = []

    def __getitem__(self, index: int) -> KeyBlob:
        return self._key_blobs[index]

    def __setitem__(self, index: int, value: KeyBlob) -> None:
        self._key_blobs.remove(self._key_blobs[index])
        self._key_blobs.insert(index, value)

    def add_key_blob(self, key_blob: KeyBlob) -> None:
        """Add key for specified address range.

        :param key_blob: to be added
        """
        self._key_blobs.append(key_blob)

    def encrypt_image(self, image: bytes, base_addr: int, byte_swap: bool) -> bytes:
        """Encrypt image with all available keyblobs.

        :param image: plain image to be encrypted
        :param base_addr: where the image will be located in target processor
        :param byte_swap: this probably depends on the flash device, how bytes are organized there
                True should be used for FLASH on EVK RT6xx; False for FLASH on EVK RT5xx
        :return: encrypted image
        """
        encrypted_data = bytearray(image)
        addr = base_addr
        for block in split_data(image, self.OTFAD_DATA_UNIT):
            for key_blob in self._key_blobs:
                if key_blob.matches_range(addr, addr + len(block)):
                    logger.debug(
                        f"Encrypting {hex(addr)}:{hex(len(block) + addr)}"
                        f" with keyblob: \n {key_blob.info()}"
                    )
                    encrypted_data[
                        addr - base_addr : len(block) + addr - base_addr
                    ] = key_blob.encrypt_image(addr, block, byte_swap, counter_value=addr)
            addr += len(block)

        return bytes(encrypted_data)

    def get_key_blobs(self) -> bytes:
        """Get key blobs.

        :return: Binary key blobs joined together
        """
        result = bytes()
        for key_blob in self._key_blobs:
            result += key_blob.plain_data()
        return align_block(
            result, 256
        )  # this is for compatibility with elftosb, probably need FLASH sector size

    def encrypt_key_blobs(
        self,
        kek: Union[bytes, str],
        key_scramble_mask: Optional[int] = None,
        key_scramble_align: Optional[int] = None,
        byte_swap_cnt: int = 0,
    ) -> bytes:
        """Encrypt key blobs with specified key.

        :param kek: key to encode key blobs
        :param key_scramble_mask: 32-bit scramble key, if KEK scrambling is desired.
        :param key_scramble_align: 8-bit scramble align, if KEK scrambling is desired.
        :param byte_swap_cnt: Encrypted keyblob reverse byte count, 0 means NO reversing is enabled
        :raises SPSDKValueError: Invalid input value.
        :return: encrypted binary key blobs joined together
        """
        if isinstance(kek, str):
            kek = bytes.fromhex(kek)
        scramble_enabled = key_scramble_mask is not None and key_scramble_align is not None
        if scramble_enabled:
            assert key_scramble_mask and key_scramble_align
            if key_scramble_mask >= 1 << 32:
                raise SPSDKValueError("OTFAD Key scramble mask has invalid length")
            if key_scramble_align >= 1 << 8:
                raise SPSDKValueError("OTFAD Key scramble align has invalid length")

            logger.debug("The scrambling of keys is enabled.")
            key_scramble_mask_inv = reverse_bits_in_bytes(
                key_scramble_mask.to_bytes(4, byteorder="big")
            )
            logger.debug(f"The inverted scramble key is: {key_scramble_mask_inv.hex()}")
        result = bytes()
        scrambled = bytes()
        for i, key_blob in enumerate(self._key_blobs):
            if scramble_enabled:
                assert key_scramble_mask and key_scramble_align
                scrambled = bytearray(kek)
                long_ix = (key_scramble_align >> (i * 2)) & 0x03
                for j in range(4):
                    scrambled[(long_ix * 4) + j] ^= key_scramble_mask_inv[j]

            logger.debug(
                f"Used KEK for keyblob{i} encryption is: {scrambled.hex() if scramble_enabled else kek.hex()}"
            )

            result += key_blob.export(
                scrambled if scramble_enabled else kek, byte_swap_cnt=byte_swap_cnt
            )
        return align_block(
            result, 256
        )  # this is for compatibility with elftosb, probably need FLASH sector size

    def info(self) -> str:
        """Text info about the instance."""
        msg = "Key-Blob\n"
        for index, key_blob in enumerate(self._key_blobs):
            msg += f"Key-Blob {str(index)}:\n"
            msg += key_blob.info()
        return msg


class OtfadNxp(Otfad):
    """OTFAD: On-the-Fly AES Decryption Module with reflecting of NXP parts."""

    def __init__(
        self,
        family: str,
        kek: Union[bytes, str],
        table_address: int = 0,
        key_blobs: Optional[List[KeyBlob]] = None,
        key_scramble_mask: Optional[int] = None,
        key_scramble_align: Optional[int] = None,
        binaries: Optional[BinaryImage] = None,
    ) -> None:
        """Constructor.

        :param family: Device family
        :param kek: KEK to encrypt OTFAD table
        :param table_address: Absolute address of OTFAD table.
        :param key_blobs: Optional Key blobs to add to OTFAD, defaults to None
        :param key_scramble_mask: If defined, the key scrambling algorithm will be applied.
            ('key_scramble_align' must be defined also)
        :param key_scramble_align: If defined, the key scrambling algorithm will be applied.
            ('key_scramble_mask' must be defined also)
        :raises SPSDKValueError: Unsupported family
        """
        super().__init__()

        if family not in self.get_supported_families():
            raise SPSDKValueError(f"Unsupported family{family} by OTFAD")

        if (key_scramble_align is None and key_scramble_mask) or (
            key_scramble_align and key_scramble_mask is None
        ):
            raise SPSDKValueError("Key Scrambling is not fully defined")

        self.family = family
        self.kek = bytes.fromhex(kek) if isinstance(kek, str) else kek
        self.key_scramble_mask = key_scramble_mask
        self.key_scramble_align = key_scramble_align
        self.table_address = table_address
        self.database = Database(OTFAD_DATABASE_FILE)
        self.blobs_min_cnt = self.database.get_device_value("key_blob_min_cnt", device=family)
        self.blobs_max_cnt = self.database.get_device_value("key_blob_max_cnt", device=family)
        self.byte_swap = self.database.get_device_value("byte_swap", device=family)
        self.key_blob_rec_size = self.database.get_device_value("key_blob_rec_size", device=family)
        self.keyblob_byte_swap_cnt = self.database.get_device_value(
            "keyblob_byte_swap_cnt", device=family
        )
        assert self.keyblob_byte_swap_cnt in [0, 2, 4, 8, 16]
        self.binaries = binaries

        if key_blobs:
            for key_blob in key_blobs:
                self.add_key_blob(key_blob)

        # Just fill up the minimum count of key blobs
        while len(self._key_blobs) < self.blobs_min_cnt:
            self.add_key_blob(
                KeyBlob(
                    start_addr=0,
                    end_addr=0,
                    key=bytes([0] * KeyBlob.KEY_SIZE),
                    counter_iv=bytes([0] * KeyBlob.CTR_SIZE),
                    key_flags=0,
                    zero_fill=bytes([0] * 4),
                )
            )

    @staticmethod
    def get_blhost_script_otp_keys(
        family: str, otp_master_key: bytes, otfad_key_seed: bytes
    ) -> str:
        """Create BLHOST script to load fuses needed to run OTFAD with OTP fuses.

        :param family: Device family.
        :param otp_master_key: OTP Master Key.
        :param otfad_key_seed: OTFAD Key Seed.
        :return: BLHOST script that loads the keys into fuses.
        """
        database = Database(OTFAD_DATABASE_FILE)
        xml_fuses = database.get_device_value("reg_fuses", device=family, default=None)
        if not xml_fuses:
            logger.debug(f"The {family} has no OTFAD fuses definition")
            return ""

        xml_fuses = os.path.join(OTFAD_DATA_FOLDER, xml_fuses)

        fuses = Registers(family, base_endianness="little")
        grouped_regs = database.get_device_value("grouped_registers", device=family, default=None)
        fuses.load_registers_from_xml(xml_fuses, grouped_regs=grouped_regs)
        reg_omk = fuses.find_reg("OTP_MASTER_KEY")
        reg_oks = fuses.find_reg("OTFAD_KEK_SEED")
        reg_omk.set_value(otp_master_key)
        reg_oks.set_value(otfad_key_seed)
        ret = (
            "# BLHOST OTFAD keys fuse programming script\n"
            f"# Generated by SPSDK {spsdk_version}\n"
            f"# Chip: {family}\n\n"
        )

        ret += f"# OTP MASTER KEY(Big Endian): {reg_omk.get_hex_value()}\n\n"
        for reg in reg_omk.sub_regs:
            ret += f"# {reg.name} fuse.\n"
            ret += f"efuse-program-once {reg.offset} 0x{reg.get_bytes_value().hex()} --no-verify\n"

        ret += f"\n# OTFAD KEK SEED (Big Endian): {reg_oks.get_hex_value()}\n\n"
        for reg in reg_oks.sub_regs:
            ret += f"# {reg.name} fuse.\n"
            ret += f"efuse-program-once {reg.offset} 0x{reg.get_bytes_value().hex()} --no-verify\n"

        return ret

    def get_blhost_script_otp_kek(self, index: int = 1) -> str:
        """Create BLHOST script to load fuses needed to run OTFAD with OTP fuses just for OTFAD key.

        :param index: Index of OTFAD peripheral [1, 2, ..., n].
        :return: BLHOST script that loads the keys into fuses.
        """
        if not self.database.get_device_value("has_kek_fuses", self.family, default=False):
            logger.debug(f"The {self.family} has no OTFAD KEK fuses")
            return ""

        peripheral_list = self.database.get_device_value("peripheral_list", device=self.family)
        if str(index) not in peripheral_list:
            logger.debug(f"The {self.family} has no OTFAD{index} peripheral")
            return ""

        filter_out_list = [f"OTFAD{i}" for i in peripheral_list if str(index) != i]
        xml_fuses = self.database.get_device_value("reg_fuses", device=self.family, default=None)
        if not xml_fuses:
            logger.debug(f"The {self.family} has no OTFAD fuses definition")
            return ""

        xml_fuses = os.path.join(OTFAD_DATA_FOLDER, xml_fuses)

        fuses = Registers(self.family, base_endianness="little")
        self.database.index = index

        grouped_regs = self.database.get_device_value(
            "grouped_registers", device=self.family, default=None
        )

        fuses.load_registers_from_xml(xml_fuses, filter_out_list, grouped_regs)

        scramble_enabled = (
            self.key_scramble_mask is not None and self.key_scramble_align is not None
        )

        otfad_key_fuse = self.database.get_device_value("otfad_key_fuse", self.family)
        otfad_cfg_fuse = self.database.get_device_value("otfad_cfg_fuse", self.family)

        fuses.find_reg(otfad_key_fuse).set_value(self.kek)
        otfad_cfg = fuses.find_reg(otfad_cfg_fuse)

        try:
            otfad_cfg.find_bitfield(
                self.database.get_device_value("otfad_enable_bitfield", self.family)
            ).set_value(1)
        except SPSDKRegsErrorBitfieldNotFound:
            logger.debug(f"Bitfield for OTFAD ENABLE not found for {self.family}")

        if scramble_enabled:
            otfad_cfg.find_bitfield(
                self.database.get_device_value("otfad_scramble_enable_bitfield", self.family)
            ).set_value(1)
            otfad_cfg.find_bitfield(
                self.database.get_device_value("otfad_scramble_align_bitfield", self.family)
            ).set_value(self.key_scramble_align)
            fuses.find_reg(
                self.database.get_device_value("otfad_scramble_key", self.family)
            ).set_value(self.key_scramble_mask)

        ret = (
            f"# BLHOST OTFAD{index} KEK fuses programming script\n"
            f"# Generated by SPSDK {spsdk_version}\n"
            f"# Chip: {self.family}, peripheral: OTFAD{index} !\n\n"
        )

        ret += f"# OTP KEK (Big Endian): {self.kek.hex()}\n\n"
        for reg in fuses.find_reg(otfad_key_fuse).sub_regs:
            ret += f"# {reg.name} fuse.\n"
            ret += f"efuse-program-once {reg.offset} 0x{reg.get_bytes_value().hex()} --no-verify\n"

        ret += f"\n\n# {otfad_cfg.name} fuse.\n"
        for bitfield in otfad_cfg.get_bitfields():
            ret += f"#   {bitfield.name}: {bitfield.get_enum_value()}\n"
        ret += f"efuse-program-once {otfad_cfg.offset} 0x{otfad_cfg.get_bytes_value().hex()} --no-verify\n"

        if scramble_enabled:
            scramble = fuses.find_reg(
                self.database.get_device_value("otfad_scramble_key", self.family)
            )
            ret += f"\n# {scramble.name} fuse.\n"
            ret += f"efuse-program-once {scramble.offset} 0x{scramble.get_bytes_value().hex()} --no-verify\n"

        return ret

    def export_image(
        self,
        plain_data: bool = False,
        swap_bytes: bool = False,
        join_sub_images: bool = True,
        table_address: int = 0,
    ) -> Optional[BinaryImage]:
        """Get the OTFAD Key Blob Binary Image representation.

        :param plain_data: Binary representation in plain data format, defaults to False
        :param swap_bytes: For some platforms the swap bytes is needed in encrypted format, defaults to False.
        :param join_sub_images: If it's True, all the binary sub-images are joined into one, defaults to True.
        :param table_address: Absolute address of OTFAD table.
        :return: OTFAD key blob data in BinaryImage.
        """
        if self.binaries is None:
            return None
        binaries: BinaryImage = deepcopy(self.binaries)
        for binary in binaries.sub_images:
            if binary.binary:
                binary.binary = align_block(binary.binary, KeyBlob._ENCRYPTION_BLOCK_SIZE)
            for segment in binary.sub_images:
                if segment.binary:
                    segment.binary = align_block(segment.binary, KeyBlob._ENCRYPTION_BLOCK_SIZE)

        binaries.validate()

        if not plain_data:
            for binary in binaries.sub_images:
                if binary.binary:
                    binary.binary = self.encrypt_image(
                        binary.binary,
                        table_address + binary.absolute_address,
                        swap_bytes,
                    )
                for segment in binary.sub_images:
                    if segment.binary:
                        segment.binary = self.encrypt_image(
                            segment.binary,
                            segment.absolute_address + table_address,
                            swap_bytes,
                        )

        if join_sub_images:
            binaries.join_images()
            binaries.validate()

        return binaries

    def binary_image(
        self,
        plain_data: bool = False,
        data_alignment: int = 16,
        otfad_table_name: str = "OTFAD_Table",
    ) -> BinaryImage:
        """Get the OTFAD Binary Image representation.

        :param plain_data: Binary representation in plain format, defaults to False
        :param data_alignment: Alignment of data part key blobs.
        :param otfad_table_name: name of the output file that contains OTFAD table
        :return: OTFAD in BinaryImage.
        """
        otfad = BinaryImage("OTFAD", offset=self.table_address)
        # Add mandatory OTFAD table
        otfad_table = (
            self.get_key_blobs()
            if plain_data
            else self.encrypt_key_blobs(
                self.kek,
                self.key_scramble_mask,
                self.key_scramble_align,
                self.keyblob_byte_swap_cnt,
            )
        )
        otfad.add_image(
            BinaryImage(
                otfad_table_name,
                size=self.key_blob_rec_size * self.blobs_max_cnt,
                offset=0,
                description=f"OTFAD description table for {self.family}",
                binary=otfad_table,
                alignment=256,
            )
        )
        binaries = self.export_image(table_address=self.table_address)

        if binaries:
            binaries.alignment = data_alignment
            binaries.validate()
            otfad.add_image(binaries)
        return otfad

    @staticmethod
    def get_supported_families() -> List[str]:
        """Get all supported families for AHAB container.

        :return: List of supported families.
        """
        database = Database(OTFAD_DATABASE_FILE)
        return database.devices.device_names

    @staticmethod
    def get_validation_schemas(family: str) -> List[Dict[str, Any]]:
        """Get list of validation schemas.

        :param family: Family for which the template should be generated.
        :return: Validation list of schemas.
        """
        if family not in OtfadNxp.get_supported_families():
            return []

        database = Database(OTFAD_DATABASE_FILE)
        schemas = ValidationSchemas.get_schema_file(OTFAD_SCH_FILE)
        family_sch = deepcopy(schemas["otfad_family"])
        family_sch["properties"]["family"]["enum"] = OtfadNxp.get_supported_families()
        family_sch["properties"]["family"]["template_value"] = family
        ret = [family_sch, schemas["otfad_output"], schemas["otfad"]]
        additional_schemes = database.get_device_value("additional_template", family, default=[])
        ret.extend([schemas[x] for x in additional_schemes])
        return ret

    @staticmethod
    def get_validation_schemas_family() -> List[Dict[str, Any]]:
        """Get list of validation schemas for family key.

        :return: Validation list of schemas.
        """
        schemas = ValidationSchemas.get_schema_file(OTFAD_SCH_FILE)
        family_sch = deepcopy(schemas["otfad_family"])
        family_sch["properties"]["family"]["enum"] = OtfadNxp.get_supported_families()
        return [family_sch]

    @staticmethod
    def generate_config_template(family: str) -> Dict[str, Any]:
        """Generate OTFAD configuration template.

        :param family: Family for which the template should be generated.
        :return: Dictionary of individual templates (key is name of template, value is template itself).
        """
        val_schemas = OtfadNxp.get_validation_schemas(family)
        database = Database(OTFAD_DATABASE_FILE)

        if val_schemas:
            template_note = database.get_device_value(
                "additional_template_text", family, default=""
            )
            title = f"On-The-Fly AES decryption Configuration template for {family}."

            yaml_data = ConfigTemplate(title, val_schemas, note=template_note).export_to_yaml()

            return {f"{family}_otfad": yaml_data}

        return {}

    @staticmethod
    def load_from_config(
        config: Dict[str, Any], config_dir: str, search_paths: Optional[List[str]] = None
    ) -> "OtfadNxp":
        """Converts the configuration option into an OTFAD image object.

        "config" content array of containers configurations.

        :param config: array of OTFAD configuration dictionaries.
        :param config_dir: directory where the config is located
        :param search_paths: List of paths where to search for the file, defaults to None
        :return: initialized OTFAD object.
        """
        otfad_config: List[Dict[str, Any]] = config["key_blobs"]
        family = config["family"]
        database = Database(OTFAD_DATABASE_FILE)
        kek = get_key(config["kek"], expected_size=16, search_paths=search_paths)
        logger.debug(f"Loaded KEK: {kek.hex()}")
        table_address = value_to_int(config["otfad_table_address"])
        start_address = min([value_to_int(addr["start_address"]) for addr in otfad_config])

        key_scramble_mask = None
        key_scramble_align = None
        if database.get_device_value("supports_key_scrambling", device=family, default=False):
            if "key_scramble" in config.keys():
                key_scramble = config["key_scramble"]
                key_scramble_mask = value_to_int(key_scramble["key_scramble_mask"])
                key_scramble_align = value_to_int(key_scramble["key_scramble_align"])

        data_blobs: Optional[List[Dict]] = config.get("data_blobs")
        binaries = None
        if data_blobs:
            # pylint: disable-next=nested-min-max
            start_address = min(
                min([value_to_int(addr["address"]) for addr in data_blobs]),
                start_address,
            )
            binaries = BinaryImage(
                filepath_from_config(
                    config, "encrypted_name", "encrypted_blobs", config_dir, config["output_folder"]
                ),
                offset=start_address - table_address,
            )
            for data_blob in data_blobs:
                data = load_binary(data_blob["data"], search_paths=search_paths)
                address = value_to_int(data_blob["address"])

                binary = BinaryImage(
                    os.path.basename(data_blob["data"]),
                    offset=address - table_address - binaries.offset,
                    binary=data,
                )
                binaries.add_image(binary)

        otfad = OtfadNxp(
            family=family,
            kek=kek,
            table_address=table_address,
            key_scramble_align=key_scramble_align,
            key_scramble_mask=key_scramble_mask,
            binaries=binaries,
        )

        for i, key_blob_cfg in enumerate(otfad_config):
            aes_key = value_to_bytes(key_blob_cfg["aes_key"], byte_cnt=KeyBlob.KEY_SIZE)
            aes_ctr = value_to_bytes(key_blob_cfg["aes_ctr"], byte_cnt=KeyBlob.CTR_SIZE)
            start_addr = value_to_int(key_blob_cfg["start_address"])
            end_addr = value_to_int(key_blob_cfg["end_address"])
            aes_decryption_enable = key_blob_cfg.get("aes_decryption_enable", True)
            valid = key_blob_cfg.get("valid", True)
            read_only = key_blob_cfg.get("read_only", True)
            flags = 0
            if aes_decryption_enable:
                flags |= KeyBlob.KEY_FLAG_ADE
            if valid:
                flags |= KeyBlob.KEY_FLAG_VLD
            if read_only:
                flags |= KeyBlob.KEY_FLAG_READ_ONLY

            otfad[i] = KeyBlob(
                start_addr=start_addr,
                end_addr=end_addr,
                key=aes_key,
                counter_iv=aes_ctr,
                key_flags=flags,
                zero_fill=bytes([0] * 4),
            )

        return otfad



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for serial communication using BUSPAL protocol, using a FRDM device as translation board."""

import datetime
import logging
import time
from enum import Enum
from typing import List, Optional, Tuple

from serial import SerialException
from serial.tools.list_ports import comports

from spsdk.mboot.exceptions import McuBootConnectionError, McuBootDataAbortError

from .base import Interface
from .uart import FPType, Uart, to_int

logger = logging.getLogger(__name__)


# pylint: disable=invalid-name
class BuspalMode(Enum):
    """Bit Bang mode command."""

    reset = 0x00  # Reset, responds "BBIO1"
    spi = 0x01  # Enter binary SPI mode, responds "SPI1"
    i2c = 0x02  # Enter binary I2C mode, responds "I2C1"


# pylint: disable=invalid-name
class BBConstants(Enum):
    """Constants."""

    reset_count = 20  # Max number of nulls to send to enter BBIO mode
    response_ok = 0x01  # Successful command response
    bulk_transfer_max = 4096  # Max number of bytes per bulk transfer
    packet_timeout_ms = 10  # Packet timeout in milliseconds


# pylint: disable=invalid-name
class Response(str, Enum):
    """Response to enter bit bang mode."""

    bitbang = "BBIO1"
    spi = "SPI1"
    i2c = "I2C1"


MODE_COMMANDS_RESPONSES = [
    {"mode": BuspalMode.reset, "response": Response.bitbang},
    {"mode": BuspalMode.spi, "response": Response.spi},
    {"mode": BuspalMode.i2c, "response": Response.i2c},
]


class Buspal(Uart):
    """BUSPAL interface."""

    BUSPAL_BAUDRATE = 57600
    DEFAULT_TIMEOUT = 5000

    mode: BuspalMode

    def __init__(self, port: str, timeout: int) -> None:
        """Initialize the BUSPAL interface.

        :param port: uart port of Buspal target
        :param timeout: timeout in milliseconds
        """
        self.timeout = timeout
        super().__init__(port, baudrate=self.BUSPAL_BAUDRATE, timeout=timeout)

    @classmethod
    def check_port(
        cls, port: str, timeout: int, props: Optional[List[str]] = None
    ) -> Optional[Interface]:
        """Check if device on comport 'port' can connect using BUSPAL communication protocol.

        :param port: name of port to check
        :param timeout: timeout in milliseconds
        :param props: buspal settings
        :return: None if device doesn't respond to PING, instance of Interface if it does
        """
        props = props if props is not None else []
        try:
            interface = cls(port=port, timeout=timeout)
            interface.open()
            interface.configure(props)
            interface.ping()
            return interface
        except (AssertionError, SerialException, McuBootConnectionError) as e:
            logger.error(str(e))
            return None

    @classmethod
    def scan_buspal(
        cls,
        port: Optional[str] = None,
        timeout: int = DEFAULT_TIMEOUT,
        props: Optional[List[str]] = None,
    ) -> List[Interface]:
        """Scan connected serial ports and set BUSPAL properties.

        Returns list of serial ports with devices that respond to BUSPAL communication protocol.
        If 'port' is specified, only that serial port is checked
        If no devices are found, return an empty list.

        :param port: name of preferred serial port, defaults to None
        :param timeout: timeout in milliseconds
        :param props: buspal target properties
        :return: list of available interfaces
        """
        if port:
            interface = cls.check_port(port, timeout, props)
            return [interface] if interface else []
        all_ports = [
            cls.check_port(comport.device, timeout, props)
            for comport in comports(include_links=True)
        ]
        return list(filter(None, all_ports))

    def open(self) -> None:
        """Open the BUSPAL interface, configure depending on selected interface."""
        self.device.open()

        # reset first, send bit-bang command
        self.enter_mode(BuspalMode.reset)
        logger.debug("Entered BB mode")
        self.enter_mode(self.mode)

    def _read_frame_header(self, expected_frame_type: Optional[int] = None) -> Tuple[int, int]:
        """Read frame header and frame type. Return them as tuple of integers.

        :param expected_frame_type: Check if the frame_type is exactly as expected
        :return: Tuple of integers representing frame header and frame type
        :raises AssertionError: Unexpected frame header or frame type (if specified)
        :raises McuBootDataAbortError: Abort frame received
        """
        header = None
        time_start = datetime.datetime.now()
        time_end = time_start + datetime.timedelta(milliseconds=self.timeout)

        # read uart until start byte is equal to FRAME_START_BYTE, max. 'retry_count' times
        while header != self.FRAME_START_BYTE and datetime.datetime.now() < time_end:
            header = to_int(self._read_default(1))
            if header == FPType.ABORT:
                raise McuBootDataAbortError()
            if header != self.FRAME_START_BYTE:
                time.sleep(BBConstants.packet_timeout_ms.value / 1000)
        assert (
            header == self.FRAME_START_BYTE
        ), f"Received invalid frame header '{header:#X}' expected '{self.FRAME_START_BYTE:#X}'"

        frame_type = to_int(self._read_default(1))

        if frame_type == FPType.ABORT:
            raise McuBootDataAbortError()
        return header, frame_type

    def configure(self, props: List[str]) -> None:
        """Configure the BUSPAL interface.

        :param props: buspal settings
        """
        raise NotImplementedError

    def enter_mode(self, mode: BuspalMode) -> None:
        """Enter BUSPAL mode.

        :param mode: buspal mode
        """
        response = next(iter(x["response"] for x in MODE_COMMANDS_RESPONSES if x["mode"] == mode))
        self._send_command_check_response(
            bytes([mode.value]), bytes(response.value.encode("utf-8"))
        )

    def _send_command_check_response(self, command: bytes, response: bytes) -> None:
        """Send a command and check if expected response is received.

        :param command: command to send
        :param response: expected response
        """
        self._write(command)
        data_recvd = self._read(len(response))
        format_received = " ".join(hex(x) for x in data_recvd)
        format_expected = " ".join(hex(x) for x in response)
        assert (
            format_received == format_expected
        ), f"Received data '{format_received}' but expected '{format_expected}'"



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2017-2018 Martin Olejar
# Copyright 2019-2022 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Commands and responses used by SDP module."""

from struct import pack, unpack_from

from spsdk.utils.easy_enum import Enum


########################################################################################################################
# SDP Command Tags
########################################################################################################################
class CommandTag(Enum):
    """SDP Commands."""

    READ_REGISTER = (0x0101, "ReadRegister", "Read data from memory or registers")
    WRITE_REGISTER = (
        0x0202,
        "WriteRegister",
        "Write one word (max 4 bytes) into memory or register",
    )
    WRITE_FILE = (0x0404, "WriteFile", "Write file (boot image) into memory")
    ERROR_STATUS = (0x0505, "ErrorStatus", "Read error code")
    WRITE_CSF = (0x0606, "WriteCsf", "Write CSF data into target")
    WRITE_DCD = (0x0A0A, "WriteDcd", "Write DCD data into target")
    SKIP_DCD_HEADER = (0x0C0C, "SkipDcdHeader", "Skip DCD content from loaded image")
    JUMP_ADDRESS = (0x0B0B, "JumpAddress", "Jump to specified address and run")
    SET_BAUDRATE = (0x0D0D, "SetBaudrate")
    PING = (0x5AA6, "Ping")


########################################################################################################################
# SDP Response Values
########################################################################################################################
class ResponseValue(Enum):
    """SDP Response Values."""

    WRITE_DATA_OK = (0x128A8A12, "Write Data Success")
    WRITE_FILE_OK = (0x88888888, "Write File Success")
    SKIP_DCD_HEADER_OK = (0x900DD009, "Skip DCD Header Success")

    LOCKED = (0x12343412, "HAB Is Enabled (Locked)")
    UNLOCKED = (0x56787856, "Hab Is Disabled (Unlocked)")

    HAB_SUCCESS = (0xF0F0F0F0, "HAB_Success", "HAB Success")
    BAUDRATE_SET = (0x09D00D90, "Baudrate Setup Success")


########################################################################################################################
# SDP Command and Response packet classes
########################################################################################################################


class CmdPacket:
    """Class representing a command packet to be sent to device."""

    FORMAT = ">HIB2IB"
    EMPTY_VALUE = 0x00

    def __init__(self, tag: CommandTag, address: int, pformat: int, count: int, value: int = 0):
        """Initialize the struct.

        :param tag: Tag number representing the command
        :param address: Address used by the command
        :param pformat: Format of the data: 8 = byte, 16 = half-word, 32 = word
        :param count: Count used by individual command
        :param value: Value to use in a particular command, defaults to 0
        """
        self.tag = tag
        self.address = address
        self.format = pformat
        self.count = count
        self.value = value

    def __str__(self) -> str:
        return "<" + self.info() + ">"

    def info(self) -> str:
        """String representation of the command packet."""
        return (
            f"Tag={CommandTag.get(self.tag, f'0x{self.tag:04X}')}, Address=0x{self.address:04X},"
            f" Format={self.format}, Count={self.count}, Value=0x{self.value:08X}"
        )

    def to_bytes(self) -> bytes:
        """Return command packet as bytes."""
        return pack(self.FORMAT, self.tag, self.address, self.format, self.count, self.value, 0)


class CmdResponse:
    """Response on the previously issued command."""

    @property
    def value(self) -> int:
        """Return a integer representation of the response."""
        return unpack_from(">I", self.raw_data)[0]

    def __init__(self, hab: bool, raw_data: bytes):
        """Initialize the response object.

        :param hab: HAB status response
        :param raw_data: Data sent by the device
        """
        self.hab = hab
        self.raw_data = raw_data

    def __str__(self) -> str:
        return "<" + self.info() + ">"

    def info(self) -> str:
        """Return stringified information about the command response."""
        return f"Response: {ResponseValue.get(self.value, f'0x{self.value:08X}')}"



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019-2023 NXP
#
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module implementing the SDPS communication protocol."""

import logging
from struct import pack
from typing import Mapping, Tuple

from spsdk.utils.misc import swap32

from .exceptions import SdpConnectionError
from .interfaces import SDPInterface

logger = logging.getLogger(__name__)

ROM_INFO = {
    "MX8QXP": {"no_cmd": True, "hid_ep1": False, "hid_pack_size": 1024},
    "MX28": {"no_cmd": False, "hid_ep1": False, "hid_pack_size": 1024},
    "MX815": {"no_cmd": True, "hid_ep1": True, "hid_pack_size": 1020},
    "MX865": {"no_cmd": True, "hid_ep1": True, "hid_pack_size": 1020},
    "MX93": {"no_cmd": True, "hid_ep1": True, "hid_pack_size": 1020},
}

BLTC_DOWNLOAD_FW = 2
CBW_BLTC_SIGNATURE = 0x43544C42

CBW_DEVICE_TO_HOST_DIR = 0x80  # "Data Out"
CBW_HOST_TO_DEVICE_DIR = 0x00  # "Data In"


class SDPS:
    """Secure Serial Downloader Protocol."""

    @property
    def name(self) -> str:
        """Get name."""
        return self.__name

    def __init__(self, device: SDPInterface, device_name: str) -> None:
        """Initialize SDPS object.

        :param device: USB device
        :param device_name: target platform name used to determine ROM settings
        """
        self._device = device
        self.__name: str = device_name

    def __enter__(self) -> "SDPS":
        self.open()
        return self

    def __exit__(self, *args: Tuple, **kwargs: Mapping) -> None:
        self.close()

    def open(self) -> None:
        """Connect to i.MX device."""
        if not self._device.is_opened:
            logger.info(f"Connect: {self._device.info()}")
            self._device.open()

    def close(self) -> None:
        """Disconnect i.MX device."""
        self._device.close()

    @property
    def is_opened(self) -> bool:
        """Indicates whether the underlying interface is open.

        :return: True if device is open, False if it's closed
        """
        return self._device.is_opened

    def write_file(self, data: bytes) -> None:
        """Write data to the target.

        :param data: The boot image data in binary format
        :raises SdpConnectionError: Timeout or Connection error
        """
        _format = "<3IB2xbI11x"

        try:
            self._device.conf(
                {
                    "hid_ep1": ROM_INFO[self.name]["hid_ep1"],
                    "pack_size": ROM_INFO[self.name]["hid_pack_size"],
                }
            )
            if not ROM_INFO[self.name]["no_cmd"]:
                cmd_packet = pack(
                    _format,
                    CBW_BLTC_SIGNATURE,
                    1,
                    len(data),
                    CBW_HOST_TO_DEVICE_DIR,
                    BLTC_DOWNLOAD_FW,
                    swap32(len(data)),
                )
                logger.info(
                    f"TX-CMD: WriteCmd(command={BLTC_DOWNLOAD_FW},"
                    f" flags=0x{CBW_HOST_TO_DEVICE_DIR:08X},"
                    f" length={len(cmd_packet)})"
                )
                self._device.write(cmd_packet)

            self._device.write(data)

        except Exception as exc:
            logger.info("RX-CMD: Timeout Error")
            raise SdpConnectionError("Timeout Error") from exc

        logger.info(f"TX-CMD: WriteFile(length={len(data)})")



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2019 Martin Olejar
# Copyright 2020-2023 NXP
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# SPDX-License-Identifier: Apache-2.0

"""EasyEnum allows numerical value, string tag and description to an enum element."""

__author__ = "Martin Olejar"
__contact__ = "martin.olejar@gmail.com"
__version__ = "0.4.0"
__license__ = "Apache 2.0"
__status__ = "Development"
__all__ = ["Enum"]

from typing import Optional, Sequence, Union

from spsdk import SPSDKError
from spsdk.exceptions import SPSDKKeyError, SPSDKTypeError

EnumKeyType = Union[str, int]


class MetaEnum(type):
    """Meta Class for Enum Type."""

    def __new__(cls, name, bases, attrs):
        _cls = super().__new__(cls, name, bases, attrs)
        _cls._items_ = []

        for base in bases:
            _cls._items_ += base._items_

        for attr, value in attrs.items():
            if attr in set(dir(type(name, (object,), {}))) or (
                attr.startswith("_") and attr.endswith("_")
            ):
                continue
            if isinstance(value, (classmethod, staticmethod)):
                continue
            if isinstance(value, tuple):
                if len(value) == 2:
                    _cls._items_.append((attr, value[0], value[1]))
                else:
                    _cls._items_.append((value[1], value[0], value[2]))
                setattr(_cls, attr, value[0])
            else:
                assert isinstance(value, int)
                _cls._items_.append((attr, value, ""))
        return _cls

    def __getitem__(cls, key):
        if isinstance(key, str):
            for name, value, _ in cls._items_:
                if key.upper() == name.upper():
                    return value
            raise SPSDKKeyError(f"'{cls.__name__}' has no item with name '{key}'")

        if isinstance(key, int):
            for name, value, _ in cls._items_:
                if key == value:
                    return name
            raise SPSDKKeyError(f"'{cls.__name__}' has no item with value '{key}'")

        raise SPSDKTypeError(f"'{cls.__name__}' has no item with type '{type(key)}'")

    def __iter__(cls):
        return (item for item in cls._items_)

    def __contains__(cls, item):
        if isinstance(item, str) and item in (item[0] for item in cls._items_):
            return True
        if isinstance(item, int) and item in (item[1] for item in cls._items_):
            return True
        return False

    def __len__(cls):
        return len(cls._items_)


class Enum(metaclass=MetaEnum):
    """Enum Type Class."""

    @classmethod
    def get(cls, key: EnumKeyType, default: Optional[EnumKeyType] = None) -> str:
        """Converts enumeration value to name OR name to enumeration value.

        :param key: either value or name (name is case INSENSITIVE)
        :param default: value in case key does not exist
        :return: name for value; value for name
        """
        try:
            return cls[key]
        except (KeyError, TypeError):
            return default

    @classmethod
    def desc(cls, key: EnumKeyType, default: str = "") -> str:
        """Description of the specified value.

        :param key: either value or name (name is case INSENSITIVE)
        :param default: value in case key does not exist
        :return: description of the value; empty string if description was not specified
        :raises SPSDKTypeError: Key is nor string or int
        """
        # pylint: disable=no-member
        if isinstance(key, str):
            for name, _, desc in cls._items_:
                if key.upper() == name.upper():
                    return desc
            return default

        if isinstance(key, int):
            for _, value, desc in cls._items_:
                if key == value:
                    return desc
            return default

        raise SPSDKTypeError(f"'{cls.__name__}' has no item with type '{type(key)}'")

    @classmethod
    def name(cls, key: int, default: Optional[str] = None) -> str:
        """Returns name of selected enumeration tag.

        :param key: enumeration tag
        :param default: value to return of tag not found; if not defined, KeyError exception will be raised
        :return: name of the corresponding enumeration tag
        :raise SPSDKKeyError: if tag not supported and default value not provided
        """
        # pylint: disable=no-member
        for name, value, _ in cls._items_:
            if key == value:
                return name

        if default is None:
            raise SPSDKKeyError("Enumeration not supported: " + str(key))

        return default

    @classmethod
    def tags(cls) -> Sequence[int]:
        """Return sequence of all enumerations tags."""
        # pylint: disable=no-member
        return [tag for _, tag, _ in cls._items_]

    @classmethod
    def from_int(cls, value: int) -> "Enum":
        """Converts integer value into enumeration.

        Note: the method does two things:
        - checks whether given integer value is defined within the enumeration
        - formally converts value into enumeration (for type hints), even the returned value is same as input
        :param value: integer value to be converted
        :return: corresponding enumeration
        :raises SPSDKError: If specified value does not match any enumeration value
        """
        if value not in cls.tags():
            raise SPSDKError(
                f"the following integer value is not defined within the enumeration: {str(value)}"
            )

        return value



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""NXP Certificate Generator."""
import logging
import os
import sys

import click
from click_option_group import optgroup

from spsdk import SPSDK_DATA_FOLDER, SPSDKError
from spsdk.apps.utils import spsdk_logger
from spsdk.apps.utils.common_cli_options import (
    CommandsTreeGroupAliasedGetCfgTemplate,
    spsdk_apps_common_options,
)
from spsdk.apps.utils.utils import SPSDKAppError, catch_spsdk_error, check_file_exists
from spsdk.crypto import (
    Encoding,
    ec,
    generate_certificate,
    load_private_key,
    load_public_key,
    save_crypto_item,
)
from spsdk.crypto.certificate_management import generate_name
from spsdk.crypto.loaders import extract_public_key, load_certificate
from spsdk.utils.misc import find_file, load_configuration, load_text, write_file

NXPCERTGEN_DATA_FOLDER: str = os.path.join(SPSDK_DATA_FOLDER, "nxpcertgen")

logger = logging.getLogger(__name__)


class CertificateParametersConfig:  # pylint: disable=too-few-public-methods
    """Configuration object for creating the certificate."""

    def __init__(self, config_data: dict) -> None:
        """Initialize cert_config from yml config data."""
        try:
            self.issuer_private_key = config_data["issuer_private_key"]
            self.subject_public_key = config_data["subject_public_key"]
            self.serial_number = config_data["serial_number"]
            self.duration = config_data["duration"]
            self.basic_constrains_ca = config_data["extensions"]["BASIC_CONSTRAINTS"]["ca"]
            self.basic_constrains_path_length = config_data["extensions"]["BASIC_CONSTRAINTS"][
                "path_length"
            ]
            self.issuer_name = generate_name(config_data["issuer"])
            self.subject_name = generate_name(config_data["subject"])
        except KeyError as e:
            raise SPSDKError(f"Error found in configuration: {e} not found") from e


@click.group(name="nxpcertgen", no_args_is_help=True, cls=CommandsTreeGroupAliasedGetCfgTemplate)  # type: ignore
@spsdk_apps_common_options
def main(log_level: int) -> None:
    """Utility for certificate generation.

    !!! The NXPCERTGEN tool is deprecated, use new NXPCRYPTO tool from SPSDK for new projects !!!
    """
    click.secho("Deprecated tool! Use npxcrypto instead", fg="yellow")
    spsdk_logger.install(level=log_level, logger=logger)


@main.command(name="generate", no_args_is_help=True)
@click.option(
    "-j",
    "-c",
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
    help="Path to yaml/json configuration file containing the parameters for certificate.",
)
@click.option(
    "-o",
    "--output",
    type=click.Path(resolve_path=True),
    required=True,
    help="Path where certificate will be stored.",
)
@click.option(
    "-e",
    "--encoding",
    required=False,
    type=click.Choice(["PEM", "DER"], case_sensitive=False),
    default="PEM",
    help="Encoding type. Default is PEM",
)
@click.option(
    "--force",
    is_flag=True,
    default=False,
    help="Force overwriting of an existing file. Create destination folder, if doesn't exist already.",
)
def generate(config: str, output: str, encoding: str, force: bool) -> None:
    """Generate certificate."""
    logger.info("Generating Certificate...")
    logger.info("Loading configuration from yml file...")

    check_file_exists(output, force)

    config_data = load_configuration(config)
    cert_config = CertificateParametersConfig(config_data)
    search_paths = [os.path.dirname(config)]

    priv_key = load_private_key(
        find_file(cert_config.issuer_private_key, search_paths=search_paths)
    )
    pub_key = load_public_key(find_file(cert_config.subject_public_key, search_paths=search_paths))

    certificate = generate_certificate(
        subject=cert_config.subject_name,
        issuer=cert_config.issuer_name,
        subject_public_key=pub_key,
        issuer_private_key=priv_key,
        serial_number=cert_config.serial_number,
        duration=cert_config.duration,
        if_ca=cert_config.basic_constrains_ca,
        path_length=cert_config.basic_constrains_path_length,
    )
    logger.info("Saving the generated certificate to the specified path...")
    encoding_type = Encoding.PEM if encoding.lower() == "pem" else Encoding.DER
    save_crypto_item(certificate, output, encoding_type=encoding_type)
    logger.info("Certificate generated successfully...")
    click.echo(f"The certificate file has been created: {output}")


@main.command(name="get-template", no_args_is_help=True)
@click.argument("output", metavar="PATH", type=click.Path(resolve_path=True))
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="Force overwriting of an existing file. Create destination folder, if doesn't exist already.",
)
def get_template(output: str, force: bool) -> None:
    """Generate the template of Certificate generation YML configuration file.

    \b
    PATH    - file name path to write template config file
    """
    logger.info("Creating Certificate template...")
    check_file_exists(output, force)

    write_file(load_text(os.path.join(NXPCERTGEN_DATA_FOLDER, "certgen_config.yaml")), output)

    click.echo(f"The configuration template file has been created: {output}")


@main.command(name="verify", no_args_is_help=True)
@click.argument("certificate", metavar="PATH", type=click.Path(exists=True, dir_okay=False))
@optgroup.group("Type of verification")
@optgroup.option(
    "-s",
    "--sign",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to key to verify certificate signature",
)
@optgroup.option(
    "-p",
    "--puk",
    type=click.Path(exists=True, dir_okay=False),
    help="Path to key to verify private key in certificate",
)
def verify(certificate: str, sign: str, puk: str) -> None:
    """Verify signature or public key in certificate.

    \b
    PATH    - path to certificate
    """
    logger.info(f"Loading certificate from: {certificate}")
    cert = load_certificate(certificate)
    if sign:
        logger.info("Performing signature verification")
        sign_algorithm = cert.signature_algorithm_oid._name
        logger.debug(f"Signature algorithm: {sign_algorithm}")
        if "ecdsa" not in sign_algorithm:
            raise SPSDKAppError(
                f"Unsupported signature algorithm: {sign_algorithm}. "
                "Only ECDSA signatures are currently supported."
            )
        verification_key = extract_public_key(sign)
        if not isinstance(verification_key, ec.EllipticCurvePublicKey):
            raise SPSDKError("Currently only ECC keys are supported.")
        if not cert.signature_hash_algorithm:
            raise SPSDKError("Certificate doesn't contain info about hashing alg.")
        try:
            verification_key.verify(
                cert.signature,
                cert.tbs_certificate_bytes,
                ec.ECDSA(cert.signature_hash_algorithm),
            )
            click.echo("Signature is OK")
        except Exception as e:
            raise SPSDKAppError("Invalid signature") from e
    if puk:
        logger.info("Performing public key verification")
        cert_puk = cert.public_key()
        if not isinstance(cert_puk, ec.EllipticCurvePublicKey):
            raise SPSDKError("Only ECC-based certificates are supported.")
        cert_puk_numbers = cert_puk.public_numbers()
        other_puk = extract_public_key(puk)
        if not isinstance(other_puk, ec.EllipticCurvePublicKey):
            raise SPSDKError("Only ECC public keys are supported")
        other_puk_numbers = other_puk.public_numbers()
        logger.debug(f"Certificate public numbers: {cert_puk_numbers}")
        logger.debug(f"Other public numbers: {other_puk_numbers}")

        if cert_puk_numbers == other_puk_numbers:
            click.echo("Public key in certificate matches the input")
        else:
            raise SPSDKAppError("Public key in certificate differs from the input")


@catch_spsdk_error
def safe_main() -> None:
    """Call the main function."""
    sys.exit(main())  # pylint: disable=no-value-for-parameter


if __name__ == "__main__":
    safe_main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module for NXP SPSDK DebugMailbox support."""

import functools
import logging
from time import sleep
from typing import Any, Dict, no_type_check

from spsdk import SPSDKError
from spsdk.debuggers.debug_probe import DebugProbe
from spsdk.exceptions import SPSDKIOError
from spsdk.utils.exceptions import SPSDKTimeoutError
from spsdk.utils.misc import Timeout

logger = logging.getLogger(__name__)


class DebugMailboxError(RuntimeError):
    """Class for DebugMailboxError."""


class DebugMailbox:
    """Class for DebugMailbox."""

    def __init__(
        self,
        debug_probe: DebugProbe,
        reset: bool = True,
        moredelay: float = 0.0,
        op_timeout: int = 1000,
    ) -> None:
        """Initialize DebugMailbox object.

        :param debug_probe: Debug probe instance.
        :param reset: Do reset of debug mailbox during initialization, defaults to True.
        :param moredelay: Time of extra delay after reset sequence, defaults to 0.0.
        :param op_timeout: Atomic operation timeout, defaults to 1000.
        :raises SPSDKIOError: Various kind of vulnerabilities during connection to debug mailbox.
        """
        # setup debug port / access point

        self.debug_probe = debug_probe
        self.dbgmlbx_ap_ix = -1
        self.reset = reset
        self.moredelay = moredelay
        # setup registers and register bitfields
        self.registers: Dict[str, Dict[str, Any]] = REGISTERS
        # set internal operation timeout
        self.op_timeout = op_timeout

        # Proceed with initiation (Resynchronization request)

        # The communication to the DM is initiated by the debugger.
        # It does so by writing the RESYNCH_REQ bit of the CSW (Control and Status Word)
        # register to 1. It then needs to reset the chip so that ROM code can observe
        # this request.
        # In order to reset the chip, the debugger can either pull the
        # reset line of the chip, or set the CHIP_RESET_REQ (This can be done at the
        # same time as setting the RESYNCH_REQ bit).

        logger.debug(f"Reset mode: {self.reset!r}")
        if self.reset:
            self.dbgmlbx_reg_write(
                addr=self.registers["CSW"]["address"],
                data=self.registers["CSW"]["bits"]["RESYNCH_REQ"]
                | self.registers["CSW"]["bits"]["CHIP_RESET_REQ"],
            )

        # Acknowledgement of initiation

        # After performing the initiation, the debugger must read back the CSW register.
        # The DM will stall the debugger until the ROM code has serviced the resynchronization request.
        # The ROM does this by performing a soft reset of the DM block, thus resetting
        # the request bit/s which were set by the debugger.
        # Therefore, the debugger must read back 0x0 in CSW to know that the initiation
        # request has been serviced.

        if self.moredelay > 0.001:
            sleep(self.moredelay)

        ret = None
        retries = 20

        while ret is None or (ret & self.registers["CSW"]["bits"]["REQ_PENDING"]):
            try:
                ret = self.dbgmlbx_reg_read(addr=self.registers["CSW"]["address"])
            except SPSDKError:
                pass
            retries -= 1
            if retries == 0:
                raise SPSDKIOError("TransferTimeoutError limit exceeded!")
            sleep(0.05)

    def read_idr(self) -> int:
        """Read IDR of debug mailbox.

        :return: IDR value of debug mailbox AP.
        """
        idr = self.dbgmlbx_reg_read(addr=REGISTERS["IDR"]["address"])
        if idr != REGISTERS["IDR"]["expected"]:
            logger.warning(
                f"The read IDR value({hex(idr)}) doesn't match the expected value: {hex(REGISTERS['IDR']['expected'])}"
            )
        return idr

    def close(self) -> None:
        """Close session."""
        self.debug_probe.close()

    def spin_read(self, reg: int) -> int:
        """Do atomic read operation to debug mailbox.

        :param reg: Register address.
        :return: Read value.
        :raises SPSDKTimeoutError: When read operation exceed defined operation timeout.
        """
        ret = None
        timeout = Timeout(self.op_timeout, units="ms")
        while ret is None:
            try:
                ret = self.dbgmlbx_reg_read(addr=reg)
            except SPSDKError as e:
                logger.debug(str(e))
                logger.debug(f"read exception  {reg:#08X}")
                if timeout.overflow():
                    raise SPSDKTimeoutError(
                        f"The Debug Mailbox read operation ends on timeout. ({str(e)})"
                    ) from e
                sleep(0.1)

        return ret

    def spin_write(self, reg: int, value: int) -> None:
        """Do atomic write operation to debug mailbox.

        :param reg: Register address.
        :param value: Value to write.
        :raises SPSDKTimeoutError: When write operation exceed defined operation timeout.
        """
        timeout = Timeout(self.op_timeout, units="ms")
        while True:
            try:
                self.dbgmlbx_reg_write(addr=reg, data=value)
                # wait for rom code to read the data
                while True:
                    ret = self.dbgmlbx_reg_read(addr=self.registers["CSW"]["address"])
                    if (ret & self.registers["CSW"]["bits"]["REQ_PENDING"]) == 0:
                        break
                    if timeout.overflow():
                        raise SPSDKTimeoutError("Mailbox command request pending timeout.")

                return
            except SPSDKError as e:
                logger.debug(str(e))
                logger.debug(f"write exception addr={reg:#08X}, val={value:#08X}")
                if timeout.overflow():
                    raise SPSDKTimeoutError(
                        f"The Debug Mailbox write operation ends on timeout. ({str(e)})"
                    ) from e
                sleep(0.1)

    @no_type_check
    # pylint: disable=no-self-argument
    def get_dbgmlbx_ap(func: Any):
        """Decorator function that secure the getting right DEBUG MAILBOX AP ix for first use.

        :param func: Decorated function.
        """
        POSSIBLE_DBGMLBX_AP_IX = [0, 2, 3]

        @functools.wraps(func)
        def wrapper(self: "DebugMailbox", *args, **kwargs):
            if self.dbgmlbx_ap_ix < 0:
                # Try to find DEBUG MAILBOX AP
                for i in POSSIBLE_DBGMLBX_AP_IX:
                    try:
                        idr = self.debug_probe.coresight_reg_read(
                            access_port=True,
                            addr=self.debug_probe.get_coresight_ap_address(
                                access_port=i, address=REGISTERS["IDR"]["address"]
                            ),
                        )
                        if idr == REGISTERS["IDR"]["expected"]:
                            self.dbgmlbx_ap_ix = i
                            logger.debug(
                                f"Found debug mailbox access port at AP{i}, IDR: 0x{idr:08X}"
                            )
                            break
                    except SPSDKError:
                        pass

                if self.dbgmlbx_ap_ix < 0:
                    raise SPSDKError("The debug mailbox access port is not found!")

            return func(self, *args, **kwargs)  # pylint: disable=not-callable

        return wrapper

    @get_dbgmlbx_ap
    def dbgmlbx_reg_read(self, addr: int = 0) -> int:
        """Read debug mailbox access port register.

        This is read debug mailbox register function for SPSDK library to support various DEBUG PROBES.

        :param addr: the register address
        :return: The read value of addressed register (4 bytes)
        :raises NotImplementedError: Derived class has to implement this method
        """
        return self.debug_probe.coresight_reg_read(
            addr=self.debug_probe.get_coresight_ap_address(
                access_port=self.dbgmlbx_ap_ix, address=addr
            )
        )

    @get_dbgmlbx_ap
    def dbgmlbx_reg_write(self, addr: int = 0, data: int = 0) -> None:
        """Write debug mailbox access port register.

        This is write debug mailbox register function for SPSDK library to support various DEBUG PROBES.

        :param addr: the register address
        :param data: the data to be written into register
        :raises NotImplementedError: Derived class has to implement this method
        """
        self.debug_probe.coresight_reg_write(
            addr=self.debug_probe.get_coresight_ap_address(
                access_port=self.dbgmlbx_ap_ix, address=addr
            ),
            data=data,
        )


REGISTERS: Dict[str, Any] = {
    # Control and Status Word (CSW) is used to control
    # the Debug Mailbox communication
    "CSW": {
        "address": 0x00,
        "bits": {
            # Debugger will set this bit to 1 to request a resynchronization
            "RESYNCH_REQ": (1 << 0),
            # Request is pending from debugger (i.e unread value in REQUEST)
            "REQ_PENDING": (1 << 1),
            # Debugger overrun error
            # (previous REQUEST overwritten before being picked up by ROM)
            "DBG_OR_ERR": (1 << 2),
            # AHB overrun Error (Return value overwritten by ROM)
            "AHB_OR_ERR": (1 << 3),
            # Soft Reset for DM (write-only from AHB,
            # not readable and self-clearing).
            # A write to this bit will cause a soft reset for DM.
            "SOFT_RESET": (1 << 4),
            # Write only bit. Once written will cause the chip to reset
            # (note that the DM is not reset by this reset as it is
            #   only resettable by a SOFT reset or a POR/BOD event)
            "CHIP_RESET_REQ": (1 << 5),
        },
    },
    # Request register is used to send data from debugger to device
    "REQUEST": {
        "address": 0x04,
    },
    # Return register is used to send data from device to debugger
    # Note: Any read from debugger side will be stalled until new data is present.
    "RETURN": {
        "address": 0x08,
    },
    # IDR register is used to identify the access port
    "IDR": {
        "address": 0xFC,
        "expected": 0x002A0000,
    },
}



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2022-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""DK6 Communication protocol."""
import logging
import struct
import time
from typing import Union

from spsdk.utils.easy_enum import Enum

from .commands import (
    CmdPacket,
    CommandTag,
    GenericResponse,
    GetChipIdResponse,
    IspUnlockResponse,
    MemBlankCheckResponse,
    MemCloseResponse,
    MemEraseResponse,
    MemGetInfoResponse,
    MemOpenResponse,
    MemoryAccessValues,
    MemoryId,
    MemReadResponse,
    MemWriteResponse,
)
from .interface import Uart

logger = logging.getLogger(__name__)


DEFAULT_KEY = b"\x11\x22\x33\x44\x55\x66\x77\x88\x11\x22\x33\x44\x55\x66\x77\x88"


class IspMode(Enum):
    """DK6 ISP modes."""

    DEFAULT = (0x00, "Default")
    START_ISP = (0x01, "Start ISP Functionality")
    UNLOCK_DEVICE = (0x7F, "Unlock device")
    EXTENDED_ISP_UNLOCK = (0x80, "Extended unlock")


class DK6Protocol:
    """Class implementing communication protocol for the DK6 devices."""

    MAX_PAYLOAD_SIZE = 512  # max size of the payload, depends on the device buffer size

    def __init__(self, device: Uart) -> None:
        """DK6Protocol constructor.

        :param device: serial device that will be used for communication.
        """
        self.uart = device

    def unlock_isp_default(self) -> IspUnlockResponse:
        """Sends unlock ISP sequence in default mode.

        It means that only Get device info command will work.

        :return: IspUnlockResponse
        """
        self.uart.write(CommandTag.UNLOCK_ISP, b"\x00")
        response = self.uart.read()
        logger.debug(response.info())

        return response

    def unlock_isp(
        self, mode: IspMode = IspMode.START_ISP, key: bytes = DEFAULT_KEY
    ) -> IspUnlockResponse:
        """Unlocks ISP with the key.

        If the key is not provided, default will be used.

        :param mode: Unlock ISP mode, defaults to IspMode.START_ISP
        :param key: default key or signed unlock key, defaults to DEFAULT_KEY
        :return: IspUnlockResponse
        """
        data = struct.pack(
            f"<B{len(key)}B",
            mode,
            *key,
        )
        packet = CmdPacket(data)
        self.uart.write(CommandTag.UNLOCK_ISP, packet)
        response = self.uart.read()
        logger.debug(response.info())

        return response

    def get_device_information(self) -> GetChipIdResponse:
        """Get device information.

        :return: GetChipIdResponse containing chip ID and chip (ROM) version
        """
        self.uart.write(CommandTag.GET_CHIPID, None)
        response = self.uart.read()
        logger.debug(response.info())

        return response

    def mem_get_info(self, memory_id: Union[MemoryId, int] = MemoryId.FLASH) -> MemGetInfoResponse:
        """Get memory info about specified memory ID.

        :param memory_id: memory ID, defaults to MemoryId.FLASH
        :return: MemGetInfoResponse containing information like size, length etc.
        """
        data = struct.pack(
            "<B",
            memory_id,
        )
        packet = CmdPacket(data)
        self.uart.write(CommandTag.MEM_GET_INFO, packet)
        response = self.uart.read()
        logger.debug(response.info())

        return response

    def mem_open(
        self,
        memory_id: MemoryId = MemoryId.FLASH,
        access: MemoryAccessValues = MemoryAccessValues.WRITE,
    ) -> MemOpenResponse:
        """Open given memory in the specified accesss mode.

        :param memory_id: memory that will be opened, defaults to MemoryId.FLASH
        :param access: access mode, defaults to MemoryAccessValues.READ
        :return: MemOpenResponse containing handle
        """
        data = struct.pack(
            "<BB",
            memory_id,
            access,
        )
        packet = CmdPacket(data)
        self.uart.write(CommandTag.MEM_OPEN, packet)
        response = self.uart.read()
        logger.debug(response.info())

        return response

    def mem_read(
        self, address: int, length: int, handle: int = 0, mode: int = 0
    ) -> MemReadResponse:
        """Read from memory.

        :param address: start address
        :param length: length of data to be read in bytes
        :param handle: handle that was returned by mem_open, defaults to 0
        :param mode: Read mode, defaults to 0
        :return: MemReadResponse containing read data
        """
        data = struct.pack("<BBII", handle, mode, address, length)
        packet = CmdPacket(data)
        self.uart.write(CommandTag.MEM_READ, packet)
        response = self.uart.read()
        logger.debug(response.info())

        return response

    def mem_write(
        self, address: int, length: int, data: bytes, handle: int = 0, mode: int = 0
    ) -> MemWriteResponse:
        """Write to memory.

        :param address: start address
        :param length: number of bytes to be written
        :param data: data to be written
        :param handle: handle returned by open memory command, defaults to 0
        :param mode: write mode, defaults to 0
        :return: MemWriteResponse
        """
        frame = struct.pack(f"<BBII{len(data)}B", handle, mode, address, length, *data)
        packet = CmdPacket(frame)

        self.uart.write(CommandTag.MEM_WRITE, packet)
        response = self.uart.read()
        logger.debug(response.info())

        return response

    def mem_close(self, handle: int = 0) -> MemCloseResponse:
        """Close the memory. Finalize writing of the memory.

        :param handle: handle returned by open memory command, defaults to 0
        :return: MemCloseResponse
        """
        self.uart.write(CommandTag.MEM_CLOSE, handle.to_bytes(1, "big"))
        response = self.uart.read()
        logger.debug(response.info())

        return response

    def mem_erase(
        self, address: int, length: int, handle: int = 0, mode: int = 0
    ) -> MemEraseResponse:
        """This command erases a region of the selected memory.

        :param address: start address
        :param length: number of bytes to be erased
        :param handle: handle returned by open memory command, defaults to 0
        :param mode: erase mode, defaults to 0
        :return: MemEraseResponse
        """
        frame = struct.pack("<BBII", handle, mode, address, length)
        packet = CmdPacket(frame)
        self.uart.write(CommandTag.MEM_ERASE, packet)
        response = self.uart.read()
        logger.debug(response.info())
        return response

    def mem_blank_check(
        self, address: int, length: int, handle: int = 0, mode: int = 0
    ) -> MemBlankCheckResponse:
        """This command checks if a region of the selected memory has been erased.

        :param address: start address
        :param length: number of bytes to be erased
        :param handle: handle returned by open memory command, defaults to 0
        :param mode: erase mode, defaults to 0
        :return: MemEraseResponse
        """
        frame = struct.pack("<BBII", handle, mode, address, length)
        packet = CmdPacket(frame)
        self.uart.write(CommandTag.MEM_BLANK_CHECK, packet)
        response = self.uart.read()
        logger.debug(response.info())

        return response

    def set_baud_rate(self, baudrate: int) -> GenericResponse:
        """Sets baudrate.

        :param baudrate: int value of baudrate to be set
        :return: GenericResponse
        """
        data = struct.pack(
            "<BI",
            0,
            baudrate,
        )
        packet = CmdPacket(data)
        self.uart.write(CommandTag.SET_BAUD, packet)
        response = self.uart.read()
        time.sleep(0.1)
        # self.uart.baudrate = baudrate
        # TODO fix baudrate change
        logger.debug(response.info())

        return response

    def reset(self) -> GenericResponse:
        """Resets device.

        :return: GenericResponse
        """
        self.uart.write(CommandTag.RESET, None)
        response = self.uart.read()
        logger.debug(response.info())

        return response

    def execute(self, address: int) -> GenericResponse:
        """This command executes (runs) code in flash or RAM.

        The response is sent before execution jumps to the provided address.

        :param address: Memory address to start execution from
        :return: GenericResponse
        """
        data = struct.pack("<I", address)
        packet = CmdPacket(data)

        self.uart.write(CommandTag.EXECUTE, packet)
        response = self.uart.read()
        logger.debug(response.info())

        return response



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Module for serial communication with a FRDM target device using BUSPAL protocol."""

import logging
import struct
from enum import Enum
from typing import Any, Dict, List, Optional

from spsdk.exceptions import SPSDKError

from .base import Interface
from .buspal import BBConstants, Buspal, BuspalMode

logger = logging.getLogger(__name__)


def scan_buspal_i2c(
    port: Optional[str] = None,
    timeout: int = Buspal.DEFAULT_TIMEOUT,
    props: Optional[List[str]] = None,
) -> List[Interface]:
    """Scan connected serial ports and set BUSPAL properties.

    Returns list of serial ports with devices that respond to BUSPAL communication protocol.
    If 'port' is specified, only that serial port is checked
    If no devices are found, return an empty list.

    :param port: name of preferred serial port, defaults to None
    :param timeout: timeout in milliseconds
    :param props: buspal target properties
    :return: list of available interfaces
    """
    return BuspalI2C.scan_buspal(port, timeout, props)


# pylint: disable=invalid-name  # changing names is fairly dangerous
class I2cModeCommand(Enum):
    """I2c mode commands."""

    exit = 0x00  # 00000000 - Exit to bit bang mode
    version = 0x01  # 00000001 - Display mode version string, responds "I2Cx"
    start_bit = 0x02  # 00000010 - I2C start bit
    stop_bit = 0x03  # 00000011 - I2C stop bit
    read_byte = 0x04  # 00000100 - I2C read byte
    ack_bit = 0x06  # 00000110 - ACK bit
    nack_bit = 0x07  # 00000111 - NACK bit
    bus_sniff = 0x0F  # 00001111 - Start bus sniffer
    bulk_write = 0x10  # 0001xxxx - Bulk I2C write, send 1-16 bytes (0=1byte!)
    configure_periph = 0x40  # 0100wxyz - Configure peripherals w=power, x=pullups, y=AUX, z=CS
    pull_up_select = 0x50  # 010100xy - Pull up voltage select (BPV4 only)- x=5v y=3.3v
    set_speed = 0x60  # 011000xx - Set I2C speed, 3=~400kHz, 2=~100kHz, 1=~50kHz, 0=~5kHz (updated in v4.2 firmware)
    set_address = 0x70  # 11100000 - Set I2C address
    write_then_read = 0x08  # Write then read


class BuspalI2C(Buspal):
    """BUSPAL I2C interface."""

    TARGET_SETTINGS = ["speed", "address"]

    HDR_FRAME_RETRY_CNT = 3

    def __init__(self, port: str, timeout: int):
        """Initialize the BUSPAL I2C interface.

        :param port: name of the serial port, defaults to None
        :param timeout: read/write timeout in milliseconds
        """
        self.mode = BuspalMode.i2c
        super().__init__(port, timeout)

    def configure(self, props: List[str]) -> None:
        """Initialize the BUSPAL I2C interface.

        :param props: buspal settings
        """
        i2c_props: Dict[str, Any] = dict(zip(self.TARGET_SETTINGS, props))

        # get I2C configuration values, use default values if settings are not defined in input string)
        speed = int(i2c_props.get("speed", 100))
        address = int(i2c_props.get("address", 0x10))

        # set I2C address
        logger.debug(f"Set I2C address to {address}")
        i2c_data = struct.pack("<BB", I2cModeCommand.set_address.value, address)
        self._send_command_check_response(i2c_data, bytes([BBConstants.response_ok.value]))

        # set I2C speed."""
        logger.debug(f"Set I2C speed to {speed}bps")
        i2c_data = struct.pack("<BI", I2cModeCommand.set_speed.value, speed)
        self._send_command_check_response(i2c_data, bytes([BBConstants.response_ok.value]))

    def _send_frame(
        self, frames: bytes, wait_for_ack: bool = True, retry_cnt: int = HDR_FRAME_RETRY_CNT
    ) -> None:
        """Send a frame to BUSPAL I2C device.

        :param frames: Data to send
        :param wait_for_ack: Wait for ACK frame from device, defaults to True
        :param retry_cnt: Number of retry in case the header frame is incorrect
        :raises AssertionError: Unexpected frame header or frame type (if specified)
        """
        size = min(len(frames), BBConstants.bulk_transfer_max.value)
        command = struct.pack("<BHH", I2cModeCommand.write_then_read.value, size, 0)
        self._write(command)
        self._send_command_check_response(frames, bytes([BBConstants.response_ok.value]))
        if wait_for_ack:
            try:
                self._read_frame_header()
            except AssertionError as error:
                # retry reading the I2C header frame in case check has failed
                if retry_cnt > 0:
                    logger.error(
                        f"{error} (retry {self.HDR_FRAME_RETRY_CNT-retry_cnt+1}/{self.HDR_FRAME_RETRY_CNT})"
                    )
                    retry_cnt -= 1
                    self._send_frame(frames, wait_for_ack, retry_cnt)
                else:
                    raise SPSDKError("Failed retrying reading the I2C header frame") from error

    def _read_default(self, size: int) -> bytes:
        """Read 'length' amount of bytes from BUSPAL I2C device.

        :return: Data read from the device
        """
        size = min(size, BBConstants.bulk_transfer_max.value)
        command = struct.pack("<BHH", I2cModeCommand.write_then_read.value, 0, size)
        self._send_command_check_response(command, bytes([BBConstants.response_ok.value]))
        return self._read(size)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Module to handle registers descriptions with support for XML files."""

import logging
import re
import xml.etree.ElementTree as ET
from typing import Any, Callable, Dict, List, Mapping, Optional, Tuple, Union
from xml.dom import minidom

from jinja2 import Environment, FileSystemLoader
from ruamel.yaml.comments import CommentedMap as CM

from spsdk import SPSDK_YML_INDENT, SPSDKError, utils
from spsdk.exceptions import SPSDKValueError
from spsdk.utils.exceptions import (
    SPSDKRegsError,
    SPSDKRegsErrorBitfieldNotFound,
    SPSDKRegsErrorEnumNotFound,
    SPSDKRegsErrorRegisterGroupMishmash,
    SPSDKRegsErrorRegisterNotFound,
)
from spsdk.utils.images import BinaryImage, BinaryPattern
from spsdk.utils.misc import format_value, value_to_bool, value_to_bytes, value_to_int, write_file

HTMLDataElement = Mapping[str, Union[str, dict, list]]
HTMLData = List[HTMLDataElement]

logger = logging.getLogger(__name__)


class RegsEnum:
    """Storage for register enumerations."""

    def __init__(self, name: str, value: Any, description: str, max_width: int = 0) -> None:
        """Constructor of RegsEnum class. Used to store enumeration information of bitfield.

        :param name: Name of enumeration.
        :param value: Value of enumeration.
        :param description: Text description of enumeration.
        :param max_width: Maximal width of enum value used to format output
        :raises SPSDKRegsError: Invalid input value.
        """
        self.name = name or "N/A"
        try:
            self.value = value_to_int(value)
        except (TypeError, ValueError, SPSDKError) as exc:
            raise SPSDKRegsError(f"Invalid Enum Value: {value}") from exc
        self.description = description or "N/A"
        self.max_width = max_width

    @classmethod
    def from_xml_element(cls, xml_element: ET.Element, maxwidth: int = 0) -> "RegsEnum":
        """Initialization Enum by XML ET element.

        :param xml_element: Input XML subelement with enumeration data.
        :param maxwidth: The maximal width of bitfield for this enum (used for formatting).
        :return: The instance of this class.
        :raises SPSDKRegsError: Error during enum XML parsing.
        """
        name = xml_element.attrib.get("name", "N/A")
        if "value" not in xml_element.attrib:
            raise SPSDKRegsError(f"Missing Enum Value Key for {name}.")

        raw_val = xml_element.attrib["value"]
        try:
            value = value_to_int(raw_val)
        except (TypeError, ValueError, SPSDKError) as exc:
            raise SPSDKRegsError(f"Invalid Enum Value: {raw_val}") from exc

        description = xml_element.attrib.get("description", "N/A")

        return cls(name, value, description, maxwidth)

    def get_value_int(self) -> int:
        """Method returns Integer value of enum.

        :return: Integer value of Enum.
        """
        return self.value

    def get_value_str(self) -> str:
        """Method returns formatted value.

        :return: Formatted string with enum value.
        """
        return format_value(self.value, self.max_width)

    def add_et_subelement(self, parent: ET.Element) -> None:
        """Creates the register XML structure in ElementTree.

        :param parent: The parent object of ElementTree.
        """
        element = ET.SubElement(parent, "bit_field_value")
        element.set("name", self.name)
        element.set("value", self.get_value_str())
        element.set("description", self.description)

    def __str__(self) -> str:
        """Overrides 'ToString()' to print register.

        :return: Friendly string with enum information.
        """
        output = ""
        output += f"Name:        {self.name}\n"
        output += f"Value:       {self.get_value_str()}\n"
        output += f"Description: {self.description}\n"

        return output


class ConfigProcessor:
    """Base class for processing configuration data."""

    NAME = "NOP"

    def __init__(self, description: str = "") -> None:
        """Initialize the processor."""
        self.description = description

    def pre_process(self, value: int) -> int:
        """Pre-process value coming from config file."""
        return value

    def post_process(self, value: int) -> int:
        """Post-process value going to config file."""
        return value

    def width_update(self, value: int) -> int:
        """Update bit-width of value going to config file."""
        return value

    @classmethod
    def get_method_name(cls, config_string: str) -> str:
        """Return config processor method name."""
        return config_string.split(":")[0]

    @classmethod
    def get_params(cls, config_string: str) -> Dict[str, int]:
        """Return config processor method parameters."""

        def split_params(param: str) -> Tuple[str, str]:
            """Split key=value pair into a tuple."""
            parts = param.split("=")
            if len(parts) != 2:
                raise SPSDKRegsError(
                    f"Invalid param setting: '{param}'. Expected format '<name>=<value>'"
                )
            return (parts[0], parts[1])

        parts = config_string.split(";", maxsplit=1)[0].split(":")
        if len(parts) == 1:
            return dict()
        params = parts[1].split(",")
        params_dict: Dict[str, str] = dict(split_params(p) for p in params)
        return {key.lower(): value_to_int(value) for key, value in params_dict.items()}

    @classmethod
    def get_description(cls, config_string: str) -> str:
        """Return extra description for config processor."""
        parts = config_string.partition(";")
        return parts[2].replace("DESC=", "")

    @classmethod
    def from_str(cls, config_string: str) -> "ConfigProcessor":
        """Create config processor instance from configuration string."""
        return cls(config_string)

    @classmethod
    def from_xml(cls, element: ET.Element) -> Optional["ConfigProcessor"]:
        """Create config processor from XML data entry."""
        processor_node = element.find("alias[@type='CONFIG_PREPROCESS']")
        if processor_node is None:
            return None
        if "value" not in processor_node.attrib:
            raise SPSDKRegsError("CONFIG_PREPROCESS alias node doesn't have a value")
        config_string = processor_node.attrib["value"]
        method_name = cls.get_method_name(config_string=config_string)
        for klass in cls.__subclasses__():
            if klass.NAME == method_name:
                return klass.from_str(config_string=config_string)
        return None


class ShiftRightConfigProcessor(ConfigProcessor):
    """Config processor performing the right-shift operation."""

    NAME = "SHIFT_RIGHT"

    def __init__(self, count: int, description: str = "") -> None:
        """Initialize the right-shift config processor.

        :param count: Count of bit for shift operation
        :param description: Extra description for config processor, defaults to ""
        """
        super().__init__(
            description=description or f"Actual binary value is shifted by {count} bits to right."
        )
        self.count = count

    def pre_process(self, value: int) -> int:
        """Pre-process value coming from config file."""
        return value >> self.count

    def post_process(self, value: int) -> int:
        """Post-process value going to config file."""
        return value << self.count

    def width_update(self, value: int) -> int:
        """Update bit-width of value going to config file."""
        return value + self.count

    @classmethod
    def from_str(cls, config_string: str) -> "ShiftRightConfigProcessor":
        """Create config processor instance from configuration string."""
        name = cls.get_method_name(config_string=config_string)
        if name != cls.NAME:
            raise SPSDKRegsError(f"Invalid method name '{name}' expected {cls.NAME}")
        params = cls.get_params(config_string=config_string)
        if "count" not in params:
            raise SPSDKRegsError(f"{cls.NAME} requires the COUNT parameter")
        description = cls.get_description(config_string=config_string)
        return cls(count=value_to_int(params["count"]), description=description)


class RegsBitField:
    """Storage for register bitfields."""

    def __init__(
        self,
        parent: "RegsRegister",
        name: str,
        offset: int,
        width: int,
        description: Optional[str] = None,
        reset_val: Any = "0",
        access: str = "RW",
        hidden: bool = False,
        config_processor: Optional[ConfigProcessor] = None,
    ) -> None:
        """Constructor of RegsBitField class. Used to store bitfield information.

        :param parent: Parent register of bitfield.
        :param name: Name of bitfield.
        :param offset: Bit offset of bitfield.
        :param width: Bit width of bitfield.
        :param description: Text description of bitfield.
        :param reset_val: Reset value of bitfield.
        :param access: Access type of bitfield.
        :param hidden: The bitfield will be hidden from standard searches.
        """
        self.parent = parent
        self.name = name or "N/A"
        self.offset = offset
        self.width = width
        self.description = description or "N/A"
        self.reset_value = value_to_int(reset_val, 0)
        self.access = access
        self.hidden = hidden
        self._enums: List[RegsEnum] = []
        self.config_processor = config_processor or ConfigProcessor()
        self.config_width = self.config_processor.width_update(width)
        self._update_reset_value()
        self.set_value(self.reset_value, raw=True)

    @classmethod
    def from_xml_element(cls, xml_element: ET.Element, parent: "RegsRegister") -> "RegsBitField":
        """Initialization register by XML ET element.

        :param xml_element: Input XML subelement with register data.
        :param parent: Reference to parent RegsRegister object.
        :return: The instance of this class.
        """
        name = xml_element.attrib.get("name", "N/A")
        offset = value_to_int(xml_element.attrib.get("offset", 0))
        width = value_to_int(xml_element.attrib.get("width", 0))
        description = xml_element.attrib.get("description", "N/A")
        access = xml_element.attrib.get("access", "R/W")
        reset_value = value_to_int(xml_element.attrib.get("reset_value", 0))
        hidden = xml_element.tag != "bit_field"
        config_processor = ConfigProcessor.from_xml(xml_element)

        bitfield = cls(
            parent, name, offset, width, description, reset_value, access, hidden, config_processor
        )

        for xml_enum in xml_element.findall("bit_field_value"):
            bitfield.add_enum(RegsEnum.from_xml_element(xml_enum, width))

        return bitfield

    def has_enums(self) -> bool:
        """Returns if the bitfields has enums.

        :return: True is has enums, False otherwise.
        """
        return len(self._enums) > 0

    def get_enums(self) -> List[RegsEnum]:
        """Returns bitfield enums.

        :return: List of bitfield enumeration values.
        """
        return self._enums

    def add_enum(self, enum: RegsEnum) -> None:
        """Add bitfield enum.

        :param enum: New enumeration value for bitfield.
        """
        self._enums.append(enum)

    def get_value(self) -> int:
        """Returns integer value of the bitfield.

        :return: Current value of bitfield.
        """
        reg_val = self.parent.get_value()
        value = reg_val >> self.offset
        mask = (1 << self.width) - 1
        value = value & mask
        value = self.config_processor.post_process(value)
        return value

    def get_reset_value(self) -> int:
        """Returns integer reset value of the bitfield.

        :return: Reset value of bitfield.
        """
        return self.reset_value

    def set_value(self, new_val: Any, raw: bool = False) -> None:
        """Updates the value of the bitfield.

        :param new_val: New value of bitfield.
        :param raw: If set, no automatic modification of value is applied.
        :raises SPSDKError: The input value is out of range.
        """
        new_val_int = value_to_int(new_val)
        new_val_int = self.config_processor.pre_process(new_val_int)
        if new_val_int > 1 << self.width:
            raise SPSDKError("The input value is out of bitfield range")
        reg_val = self.parent.get_value()
        mask = ((1 << self.width) - 1) << self.offset
        reg_val = reg_val & ~mask
        value = (new_val_int << self.offset) & mask
        reg_val = reg_val | value
        self.parent.set_value(reg_val, raw)

    def _update_reset_value(self) -> None:
        """Updates the reset value of the bitfield in register."""
        reg_val = self.parent.get_reset_value()
        mask = ((1 << self.width) - 1) << self.offset
        reg_val = reg_val & ~mask
        value = (self.reset_value << self.offset) & mask
        reg_val = reg_val | value
        self.parent._reset_value = reg_val  # pylint: disable=protected-access

    def set_enum_value(self, new_val: str, raw: bool = False) -> None:
        """Updates the value of the bitfield by its enum value.

        :param new_val: New enum value of bitfield.
        :param raw: If set, no automatic modification of value is applied.
        :raises SPSDKRegsErrorEnumNotFound: Input value cannot be decoded.
        """
        try:
            val_int = self.get_enum_constant(new_val)
        except SPSDKRegsErrorEnumNotFound:
            # Try to decode standard input
            try:
                val_int = value_to_int(new_val)
            except TypeError:
                raise SPSDKRegsErrorEnumNotFound  # pylint: disable=raise-missing-from
        self.set_value(val_int, raw)

    def get_enum_value(self) -> Union[str, int]:
        """Returns enum value of the bitfield.

        :return: Current value of bitfield.
        """
        value = self.get_value()
        for enum in self._enums:
            if enum.get_value_int() == value:
                return enum.name
        # return value
        return self.get_hex_value()

    def get_hex_value(self) -> str:
        """Get the value of register in string hex format.

        :return: Hexadecimal value of register.
        """
        fmt = f"0{self.config_width // 4}X"
        val = f"0x{format(self.get_value(), fmt)}"
        return val

    def get_enum_constant(self, enum_name: str) -> int:
        """Returns constant representation of enum by its name.

        :return: Constant of enum.
        :raises SPSDKRegsErrorEnumNotFound: The enum has not been found.
        """
        for enum in self._enums:
            if enum.name == enum_name:
                return enum.get_value_int()

        raise SPSDKRegsErrorEnumNotFound(f"The enum for {enum_name} has not been found.")

    def get_enum_names(self) -> List[str]:
        """Returns list of the enum strings.

        :return: List of enum names.
        """
        return [x.name for x in self._enums]

    def add_et_subelement(self, parent: ET.Element) -> None:
        """Creates the register XML structure in ElementTree.

        :param parent: The parent object of ElementTree.
        """
        element = ET.SubElement(parent, "reserved_bit_field" if self.hidden else "bit_field")
        element.set("offset", hex(self.offset))
        element.set("width", str(self.width))
        element.set("name", self.name)
        element.set("access", self.access)
        element.set("reset_value", format_value(self.reset_value, self.width))
        element.set("description", self.description)
        for enum in self._enums:
            enum.add_et_subelement(element)

    def get_html_data_element(self) -> HTMLDataElement:
        """Returns HTML element of bitfield.

        :return: HTML element.
        """
        enum_desc = {}
        for enum in self.get_enums():
            enum_desc[enum.get_value_str()] = enum.description

        return {
            "name": self.name,
            "desc": self.description,
            "width": str(self.width),
            "offset": str(self.offset),
            "bit_values": enum_desc,
        }

    def __str__(self) -> str:
        """Override 'ToString()' to print register.

        :return: Friendly looking string that describes the bitfield.
        """
        output = ""
        output += f"Name:     {self.name}\n"
        output += f"Offset:   {self.offset} bits\n"
        output += f"Width:    {self.width} bits\n"
        output += f"Access:   {self.access} bits\n"
        output += f"Reset val:{self.reset_value}\n"
        output += f"Description: \n {self.description}\n"
        if self.hidden:
            output += "This is hidden bitfield!\n"

        i = 0
        for enum in self._enums:
            output += f"Enum             #{i}: \n" + str(enum)
            i += 1

        return output


class RegsRegister:
    """Initialization register by input information."""

    def __init__(
        self,
        name: str,
        offset: int,
        width: int,
        description: Optional[str] = None,
        reverse: bool = False,
        access: Optional[str] = None,
        config_as_hexstring: bool = False,
        otp_index: Optional[int] = None,
        reverse_subregs_order: bool = False,
        base_endianness: str = "big",
    ) -> None:
        """Constructor of RegsRegister class. Used to store register information.

        :param name: Name of register.
        :param offset: Byte offset of register.
        :param width: Bit width of register.
        :param description: Text description of register.
        :param reverse: Multi byte register value could be printed in reverse order.
        :param access: Access type of register.
        :param config_as_hexstring: Config is stored as a hex string.
        :param otp_index: Index of OTP fuse.
        :param reverse_subregs_order: Reverse order of sub registers.
        :param base_endianness: Base endianness for bytes import/export of value.
        """
        self.name = name
        self.offset = offset
        self.width = width
        self.description = description or "N/A"
        self.access = access or "RW"
        self.reverse = reverse
        self._bitfields: List[RegsBitField] = []
        self._set_value_hooks: list = []
        self._value = 0
        self._reset_value = 0
        self.config_as_hexstring = config_as_hexstring
        self.otp_index = otp_index
        self.reverse_subregs_order = reverse_subregs_order
        self.base_endianness = base_endianness

        # Grouped register members
        self.sub_regs: List["RegsRegister"] = []
        self._sub_regs_width_init = False
        self._sub_regs_width = 0

    @classmethod
    def from_xml_element(cls, xml_element: ET.Element) -> "RegsRegister":
        """Initialization register by XML ET element.

        :param xml_element: Input XML subelement with register data.
        :return: The instance of this class.
        """
        name = xml_element.attrib.get("name", "N/A")
        offset = value_to_int(xml_element.attrib.get("offset", 0))
        width = value_to_int(xml_element.attrib.get("width", 0))
        description = xml_element.attrib.get("description", "N/A")
        reverse_subregs_order = (xml_element.attrib.get("reverse_subregs_order", "False")) == "True"
        reverse = (xml_element.attrib.get("reversed", "False")) == "True"
        access = xml_element.attrib.get("access", "N/A")
        otp_index_raw = xml_element.attrib.get("otp_index")
        otp_index = None
        if otp_index_raw:
            otp_index = value_to_int(otp_index_raw)
        reg = cls(
            name,
            offset,
            width,
            description,
            reverse,
            access,
            otp_index=otp_index,
            reverse_subregs_order=reverse_subregs_order,
        )
        value = xml_element.attrib.get("value")
        if value:
            reg.set_value(value)

        if xml_element.text:
            xml_bitfields = xml_element.findall("bit_field")
            xml_bitfields.extend(xml_element.findall("reserved_bit_field"))
            xml_bitfields_len = len(xml_bitfields)
            for xml_bitfield in xml_bitfields:
                bitfield = RegsBitField.from_xml_element(xml_bitfield, reg)
                if xml_bitfields_len == 1 and bitfield.width == reg.width:
                    if len(reg.description) < len(bitfield.description):
                        reg.description = bitfield.description
                    reg.access = bitfield.access
                else:
                    if reg.access == "N/A":
                        reg.access = "Bitfields depended"
                    reg.add_bitfield(bitfield)
        return reg

    def has_group_registers(self) -> bool:
        """Returns true if register is compounded from sub-registers.

        :return: True if register has sub-registers, False otherwise.
        """
        return len(self.sub_regs) > 0

    def add_group_reg(self, reg: "RegsRegister") -> None:
        """Add group element for this register.

        :param reg: Register member of this register group.
        :raises SPSDKRegsErrorRegisterGroupMishmash: When any inconsistency is detected.
        """
        first_member = not self.has_group_registers()
        if first_member:
            if self.offset == 0:
                self.offset = reg.offset
            if self.width == 0:
                self.width = reg.width
            else:
                self._sub_regs_width_init = True
                self._sub_regs_width = reg.width
            if self.access == "RW":
                self.access = reg.access
        else:
            # There is strong rule that supported group MUST be in one row in memory!
            if not self._sub_regs_width_init:
                if self.offset + self.width // 8 != reg.offset:
                    raise SPSDKRegsErrorRegisterGroupMishmash(
                        f"The register {reg.name} doesn't follow the previous one."
                    )
                self.width += reg.width
            else:
                if self.offset + self.width // 8 <= reg.offset:
                    raise SPSDKRegsErrorRegisterGroupMishmash(
                        f"The register {reg.name} doesn't follow the previous one."
                    )
                self._sub_regs_width += reg.width
                if self._sub_regs_width > self.width:
                    raise SPSDKRegsErrorRegisterGroupMishmash(
                        f"The register {reg.name} bigger width than is defined."
                    )
            if self.sub_regs[0].width != reg.width:
                raise SPSDKRegsErrorRegisterGroupMishmash(
                    f"The register {reg.name} has different width."
                )
            if self.access != reg.access:
                raise SPSDKRegsErrorRegisterGroupMishmash(
                    f"The register {reg.name} has different access type."
                )
        reg.base_endianness = self.base_endianness
        self.sub_regs.append(reg)

    def add_et_subelement(self, parent: ET.Element) -> None:
        """Creates the register XML structure in ElementTree.

        :param parent: The parent object of ElementTree.
        """
        element = ET.SubElement(parent, "register")
        element.set("offset", hex(self.offset))
        element.set("width", str(self.width))
        element.set("name", self.name)
        element.set("reversed", str(self.reverse))
        element.set("description", self.description)
        if self.otp_index:
            element.set("otp_index", str(self.otp_index))
        for bitfield in self._bitfields:
            bitfield.add_et_subelement(element)

    def get_html_data_element(self, exclude: Optional[List[str]] = None) -> HTMLDataElement:
        """Returns HTML element of register.

        :return: HTML element.
        """
        bitfield_desc = []
        for bitfield in self.get_bitfields(exclude):
            bitfield_desc.append(bitfield.get_html_data_element())

        return {
            "name": self.name,
            "desc": self.description,
            "width": str(self.width),
            "offset": hex(self.offset),
            "bitfields": bitfield_desc,
        }

    def set_value(self, val: Any, raw: bool = False) -> None:
        """Set the new value of register.

        :param val: The new value to set.
        :param raw: Do not use any modification hooks.
        :raises SPSDKError: When invalid values is loaded into register
        """
        try:
            value = value_to_int(val)
            if value >= 1 << self.width:
                raise SPSDKError(
                    f"Input value {value} doesn't fit into register of width {self.width}."
                )
            if not raw and len(self._set_value_hooks) > 0:
                for hook in self._set_value_hooks:
                    value = hook[0](value, hook[1])

            self._value = value

            if self.has_group_registers():
                # Update also values in sub registers
                subreg_width = self.sub_regs[0].width
                for index, sub_reg in enumerate(self.sub_regs, start=1):
                    if self.reverse_subregs_order:
                        bit_pos = (index - 1) * subreg_width
                    else:
                        bit_pos = self.width - index * subreg_width
                    sub_reg.set_value((value >> bit_pos) & ((1 << subreg_width) - 1), raw=raw)

        except SPSDKError as exc:
            raise SPSDKError(f"Loaded invalid value {str(val)}") from exc

    def reset_value(self, raw: bool = False) -> None:
        """Reset the value of register.

        :param raw: Do not use any modification hooks.
        """
        value = 0
        for bitfield in self.get_bitfields():
            width = bitfield.width
            offset = bitfield.offset
            val = bitfield.reset_value
            value |= (val & ((1 << width) - 1)) << offset

        self.set_value(value, raw)

    def get_value(self) -> int:
        """Get the value of register."""
        if self.has_group_registers():
            # Update local value, by the sub register values
            subreg_width = self.sub_regs[0].width
            sub_regs_value = 0
            for index, sub_reg in enumerate(self.sub_regs, start=1):
                if self.reverse_subregs_order:
                    bit_pos = (index - 1) * subreg_width
                else:
                    bit_pos = self.width - index * subreg_width
                sub_regs_value |= sub_reg.get_value() << (bit_pos)
            if sub_regs_value != self._value:
                self.set_value(sub_regs_value, raw=True)

        return self._value

    def get_bytes_value(self, reverse_off: bool = False) -> bytes:
        """Get the bytes value of register.

        :param reverse_off: The value indianness is returned by 'reversed' member if True,
            the reverse member is ignored otherwise.
        :return: Register value in bytes.
        """
        rev_endianness = "little" if self.base_endianness == "big" else "big"
        endianness = self.base_endianness if (not self.reverse or reverse_off) else rev_endianness
        return value_to_bytes(
            self.get_value(),
            align_to_2n=False,
            byte_cnt=self.width // 8,
            endianness=endianness,  # type: ignore[arg-type]
        )

    def set_bytes_value(self, value: bytes, reverse_off: bool = False) -> None:
        """Set register value from the bytes value.

        :param value: The input value in raw bytes.
        :param reverse_off: The value indianness is loaded by 'reversed' member if True,
            the reverse member is ignored otherwise.
        :raises SPSDKValueError: Invalid input value.
        """
        if len(value) < self.width // 8:
            raise SPSDKValueError(f"Invalid length of inputs bytes to set {self.name} register.")
        rev_endianness = "little" if self.base_endianness == "big" else "big"
        endianness = self.base_endianness if (not self.reverse or reverse_off) else rev_endianness
        self.set_value(
            int.from_bytes(value[: self.width // 8], byteorder=endianness),  # type: ignore[arg-type]
            raw=True,
        )

    def get_hex_value(self) -> str:
        """Get the value of register in string hex format.

        :return: Hexadecimal value of register.
        """
        fmt = f"0{self.width // 4}X"
        val = f"{'' if self.config_as_hexstring else '0x'}{format(self.get_value(), fmt)}"
        return val

    def get_reset_value(self) -> int:
        """Returns reset value of the register.

        :return: Reset value of register.
        """
        return self._reset_value

    def add_bitfield(self, bitfield: RegsBitField) -> None:
        """Add register bitfield.

        :param bitfield: New bitfield value for register.
        """
        self._bitfields.append(bitfield)

    def get_bitfields(self, exclude: Optional[List[str]] = None) -> List[RegsBitField]:
        """Returns register bitfields.

        Method allows exclude some bitfields by their names.
        :param exclude: Exclude list of bitfield names if needed.
        :return: Returns List of register bitfields.
        """
        ret = []
        for bitf in self._bitfields:
            if bitf.hidden:
                continue
            if exclude and bitf.name.startswith(tuple(exclude)):
                continue
            ret.append(bitf)
        return ret

    def get_bitfield_names(self, exclude: Optional[List[str]] = None) -> List[str]:
        """Returns list of the bitfield names.

        :param exclude: Exclude list of bitfield names if needed.
        :return: List of bitfield names.
        """
        return [x.name for x in self.get_bitfields(exclude)]

    def find_bitfield(self, name: str) -> RegsBitField:
        """Returns the instance of the bitfield by its name.

        :param name: The name of the bitfield.
        :return: Instance of the bitfield.
        :raises SPSDKRegsErrorBitfieldNotFound: The bitfield doesn't exist.
        """
        for bitfield in self._bitfields:
            if name == bitfield.name:
                return bitfield

        raise SPSDKRegsErrorBitfieldNotFound(f" The {name} is not found in register {self.name}.")

    def add_setvalue_hook(self, hook: Callable, context: Optional[Any] = None) -> None:
        """Set the value hook for write operation.

        :param hook: Callable hook for set value operation.
        :param context: Context data for this hook.
        """
        self._set_value_hooks.append((hook, context))

    def __str__(self) -> str:
        """Override 'ToString()' to print register.

        :return: Friendly looking string that describes the register.
        """
        output = ""
        output += f"Name:   {self.name}\n"
        output += f"Offset: 0x{self.offset:04X}\n"
        output += f"Width:  {self.width} bits\n"
        output += f"Access:   {self.access}\n"
        output += f"Description: \n {self.description}\n"
        if self.otp_index:
            output += f"OTP Word: \n {self.otp_index}\n"

        i = 0
        for bitfield in self._bitfields:
            output += f"Bitfield #{i}: \n" + str(bitfield)
            i += 1

        return output


class Registers:
    """SPSDK Class for registers handling."""

    def __init__(self, device_name: str, base_endianness: str = "big") -> None:
        """Initialization of Registers class."""
        assert base_endianness in ["little", "big"]
        self._registers: List[RegsRegister] = []
        self.dev_name = device_name
        self.base_endianness = base_endianness

    def find_reg(self, name: str, include_group_regs: bool = False) -> RegsRegister:
        """Returns the instance of the register by its name.

        :param name: The name of the register.
        :param include_group_regs: The algorithm will check also group registers.
        :return: Instance of the register.
        :raises SPSDKRegsErrorRegisterNotFound: The register doesn't exist.
        """
        for reg in self._registers:
            if name == reg.name:
                return reg
            if include_group_regs and reg.has_group_registers():
                for sub_reg in reg.sub_regs:
                    if name == sub_reg.name:
                        return sub_reg

        raise SPSDKRegsErrorRegisterNotFound(
            f"The {name} is not found in loaded registers for {self.dev_name} device."
        )

    def add_register(self, reg: RegsRegister) -> None:
        """Adds register into register list.

        :param reg: Register to add to the class.
        :raises SPSDKError: Invalid type has been provided.
        :raise SPSDKRegsError: Cannot add register with same name
        """
        if not isinstance(reg, RegsRegister):
            raise SPSDKError("The 'reg' has invalid type.")

        if reg.name in self.get_reg_names():
            raise SPSDKRegsError(f"Cannot add register with same name: {reg.name}.")

        # update base endianness for all registers in group
        reg.base_endianness = self.base_endianness
        self._registers.append(reg)

    def remove_register(self, reg: RegsRegister) -> None:
        """Remove register from register list by its instance reference.

        :reg: Instance of register that should be removed.
        :raises SPSDKError: Invalid type has been provided.
        """
        if not isinstance(reg, RegsRegister):
            raise SPSDKError("The 'reg' has invalid type.")

        self._registers.remove(reg)

    def remove_register_by_name(self, reg_names: List[str]) -> None:
        """Removes register from register list by List of its names.

        :reg_names: List of names of registers that should be removed.
        """
        for reg in self._registers:
            if any(reg.name in name for name in reg_names):
                self._registers.remove(reg)

    def get_registers(
        self, exclude: Optional[List[str]] = None, include_group_regs: bool = False
    ) -> List[RegsRegister]:
        """Returns list of the registers.

        Method allows exclude some register by their names.
        :param exclude: Exclude list of register names if needed.
        :param include_group_regs: The algorithm will check also group registers.
        :return: List of register names.
        """
        if exclude:
            regs = [r for r in self._registers if not r.name.startswith(tuple(exclude))]
        else:
            regs = self._registers.copy()
        if include_group_regs:
            sub_regs = []
            for reg in regs:
                if reg.has_group_registers():
                    sub_regs.extend(reg.sub_regs)
            regs.extend(sub_regs)

        return regs

    def get_reg_names(
        self, exclude: Optional[List[str]] = None, include_group_regs: bool = False
    ) -> List[str]:
        """Returns list of the register names.

        :param exclude: Exclude list of register names if needed.
        :param include_group_regs: The algorithm will check also group registers.
        :return: List of register names.
        """
        return [x.name for x in self.get_registers(exclude, include_group_regs)]

    def run_hooks(self, exclude: Optional[List[str]] = None) -> None:
        """The method run hooks on all regular registers.

        :param exclude: The list of register names to be excluded.
        """
        for reg in self.get_registers(exclude):
            reg.set_value(reg.get_value(), False)

    def reset_values(self, exclude: Optional[List[str]] = None) -> None:
        """The method reset values in registers.

        :param exclude: The list of register names to be excluded.
        """
        for reg in self.get_registers(exclude):
            reg.reset_value(True)

    def __str__(self) -> str:
        """Override 'ToString()' to print register.

        :return: Friendly looking string that describes the registers.
        """
        output = ""
        output += "Device name:        " + self.dev_name + "\n"
        for reg in self._registers:
            output += str(reg) + "\n"

        return output

    def write_xml(self, file_name: str) -> None:
        """Write loaded register structures into XML file.

        :param file_name: The name of XML file that should be created.
        """
        xml_root = ET.Element("regs")
        for reg in self._registers:
            reg.add_et_subelement(xml_root)

        no_pretty_data = minidom.parseString(
            ET.tostring(xml_root, encoding="unicode", short_empty_elements=False)
        )
        write_file(no_pretty_data.toprettyxml(), file_name, encoding="utf-8")

    def image_info(
        self, size: int = 0, pattern: BinaryPattern = BinaryPattern("zeros")
    ) -> BinaryImage:
        """Export Registers into  binary information.

        :param size: Result size of Image, 0 means automatic minimal size.
        :param pattern: Pattern of gaps, defaults to "zeros"
        """
        image = BinaryImage(self.dev_name, size=size, pattern=pattern)
        for reg in self._registers:
            image.add_image(
                BinaryImage(
                    reg.name,
                    reg.width // 8,
                    offset=reg.offset // 8,
                    description=reg.description,
                    binary=reg.get_bytes_value(),
                )
            )

        return image

    def export(self, size: int = 0, pattern: BinaryPattern = BinaryPattern("zeros")) -> bytes:
        """Export Registers into binary.

        :param size: Result size of Image, 0 means automatic minimal size.
        :param pattern: Pattern of gaps, defaults to "zeros"
        """
        return self.image_info(size, pattern).export()

    def get_validation_schema(self) -> Dict:
        """Get the JSON SCHEMA for registers.

        :return: JSON SCHEMA.
        """
        properties = {}
        for reg in self.get_registers():
            bitfields = reg.get_bitfields()
            if bitfields:
                bitfields_schema = {}
                for bitfield in bitfields:
                    bitfields_schema[bitfield.name] = {
                        "type": ["string", "number"],
                        "title": f"{bitfield.name}",
                        "description": f"{bitfield.description}",
                        "template_value": bitfield.get_value(),
                    }
                properties[reg.name] = {
                    "type": "object",
                    "title": f"{reg.name}",
                    "description": f"{reg.description}",
                    "required": ["bitfields"],
                    "properties": {"bitfields": {"type": "object", "properties": bitfields_schema}},
                }
            else:
                properties[reg.name] = {
                    "type": "object",
                    "title": f"{reg.name}",
                    "description": f"{reg.description}",
                    "required": ["value"],
                    "properties": {
                        "value": {
                            "type": ["string", "number"],
                            "title": f"{reg.name}",
                            "description": f"{reg.description}",
                            "format": "number",
                            "template_value": f"{reg.get_hex_value()}",
                        }
                    },
                }

        return {"type": "object", "title": self.dev_name, "properties": properties}

    def generate_html(
        self,
        heading1: str,
        heading2: str,
        regs_exclude: Optional[List[str]] = None,
        fields_exclude: Optional[List[str]] = None,
    ) -> str:
        """Generate describing HTML file with registers.

        :param heading1: The main title in HTML.
        :param heading2: The sub-title in HTML.
        :param regs_exclude: The exclude registers list.
        :param fields_exclude: The exclude bitfield list.
        :return: The content of HTML file.
        """
        data: HTMLData = []
        for reg in self.get_registers(regs_exclude):
            data.append(reg.get_html_data_element(fields_exclude))

        jinja_env = Environment(loader=FileSystemLoader(utils.REGS_DATA_FOLDER), autoescape=True)
        template = jinja_env.get_template("regs_desc_template.html")
        return template.render(heading1=heading1, heading2=heading2, data=data)

    # pylint: disable=no-self-use   #It's better to have this function visually close to callies
    def _filter_by_names(self, items: List[ET.Element], names: List[str]) -> List[ET.Element]:
        """Filter out all items in the "items" tree,whose name starts with one of the strings in "names" list.

        :param items: Items to be filtered out.
        :param names: Names to filter out.
        :return: Filtered item elements list.
        """
        return [item for item in items if not item.attrib["name"].startswith(tuple(names))]

    # pylint: disable=dangerous-default-value
    def load_registers_from_xml(
        self,
        xml: str,
        filter_reg: Optional[List[str]] = None,
        grouped_regs: Optional[List[dict]] = None,
    ) -> None:
        """Function loads the registers from the given XML.

        :param xml: Input XML data in string format.
        :param filter_reg: List of register names that should be filtered out.
        :param grouped_regs: List of register prefixes names to be grouped into one.
        :raises SPSDKRegsError: XML parse problem occurs.
        """

        def is_reg_in_group(reg: str) -> Union[dict, None]:
            """Help function to recognize if the register should be part of group."""
            if grouped_regs:
                for group in grouped_regs:
                    # pylint: disable=anomalous-backslash-in-string  # \d is a part of the regex pattern
                    if re.fullmatch(f"{group['name']}\d+", reg) is not None:
                        return group
            return None

        try:
            xml_elements = ET.parse(xml)
        except ET.ParseError as exc:
            raise SPSDKRegsError(f"Cannot Parse XML data: {str(exc)}") from exc
        xml_registers = xml_elements.findall("register")
        xml_registers = self._filter_by_names(xml_registers, filter_reg or [])
        # Load all registers into the class
        for xml_reg in xml_registers:
            group = is_reg_in_group(xml_reg.attrib["name"])
            if group:
                try:
                    group_reg = self.find_reg(group["name"])
                except SPSDKRegsErrorRegisterNotFound:
                    group_reg = RegsRegister(
                        name=group["name"],
                        offset=value_to_int(group.get("offset", 0)),
                        width=value_to_int(group.get("width", 0)),
                        description=group.get(
                            "description", f"Group of {group['name']} registers."
                        ),
                        reverse=value_to_bool(group.get("reversed", False)),
                        access=group.get("access", None),
                        config_as_hexstring=group.get("config_as_hexstring", False),
                        reverse_subregs_order=group.get("reverse_subregs_order", False),
                    )

                    self.add_register(group_reg)
                group_reg.add_group_reg(RegsRegister.from_xml_element(xml_reg))
            else:
                self.add_register(RegsRegister.from_xml_element(xml_reg))

    def load_yml_config(
        self,
        yml_data: Any,
        exclude_regs: Optional[List[str]] = None,
        exclude_fields: Optional[Dict[str, Dict[str, str]]] = None,
    ) -> None:
        """The function loads the configuration from YML file.

        :param yml_data: The YAML commented data with register values.
        :param exclude_regs: List of excluded registers
        :param exclude_fields: Dictionary with lists of excluded bitfields
        """
        for reg_name in yml_data.keys():
            if exclude_regs and reg_name in exclude_regs:
                continue
            reg_dict = yml_data[reg_name]
            register = self.find_reg(reg_name, include_group_regs=True)
            if "value" in reg_dict.keys():
                raw_val = reg_dict["value"]
                val = (
                    int(raw_val, 16)
                    if register.config_as_hexstring and isinstance(raw_val, str)
                    else value_to_int(raw_val)
                )
                register.set_value(val, True)
            elif "bitfields" in reg_dict.keys():
                for bitfield_name in reg_dict["bitfields"]:
                    bitfield_val = reg_dict["bitfields"][bitfield_name]
                    try:
                        bitfield = register.find_bitfield(bitfield_name)
                    except SPSDKRegsErrorBitfieldNotFound:
                        logger.error(
                            f"The {bitfield_name} is not found in register {register.name}."
                        )
                        continue
                    if (
                        exclude_fields
                        and reg_name in exclude_fields.keys()
                        and bitfield_name in exclude_fields[reg_name]
                    ):
                        continue

                    try:
                        bitfield.set_enum_value(bitfield_val, True)
                    except SPSDKError:
                        # New versions of register data do not contain register and bitfield value in enum
                        old_bitfield = bitfield_val
                        bitfield_val = bitfield_val.replace(bitfield.name + "_", "").replace(
                            register.name + "_", ""
                        )
                        # Some bitfield were renamed from ENABLE to ALLOW
                        bitfield_val = "ALLOW" if bitfield_val == "ENABLE" else bitfield_val
                        logger.warning(
                            f"Bitfield {old_bitfield} not found, trying backward compatibility mode with {bitfield_val}"
                        )
                        bitfield.set_enum_value(bitfield_val, True)
            else:
                logger.error(f"There are no data for {reg_name} register.")

            logger.debug(f"The register {reg_name} has been loaded from configuration.")

    def _get_bitfield_yaml_description(self, bitfield: RegsBitField, indent: int) -> str:
        """Create the valuable comment for bitfield.

        :param bitfield: Bitfield used to generate description.
        :param indent: Indent for multiline comment.
        :return: Bitfield YAML description.
        """

        def new_line(comment: str) -> str:
            return f"\n{' '*indent}# {comment}"

        description = f"Width: {bitfield.config_width}b"
        if bitfield.description not in ("", "."):
            description += ", Description: " + bitfield.description.replace(
                "&#10;", f"\n{' '*indent}# "
            )
        if bitfield.config_processor.description:
            description += ". NOTE: " + bitfield.config_processor.description
        if bitfield.has_enums():
            for enum in bitfield.get_enums():
                descr = enum.description if enum.description != "." else enum.name
                enum_description = descr.replace("&#10;", f"\n{' '*indent}# ")
                description += new_line(
                    f"- {enum.name}, ({enum.get_value_int()}): {enum_description}"
                )
        return description

    def create_yml_config_white_list(
        self, white_list: Optional[Dict[str, Any]] = None, diff: bool = False, indent: int = 0
    ) -> Any:
        """The function creates the configuration YML file.

        :param white_list: Dictionary with lists of registers and its bitfields
        :param diff: Get only configuration with difference value to reset state.
        :param indent: Indent in space to generate YML.
        :return: YAML commented map of registers value.
        """
        data = CM()
        index = 0
        for reg in self.get_registers():
            if white_list and reg.name not in white_list.keys():
                continue
            reg_yml = CM()
            if diff and reg.get_value() == reg.get_reset_value():
                continue
            descr = reg.description if reg.description != "." else reg.name
            descr = descr.replace("&#10;", "\n# ")
            data.insert(
                index,
                reg.name,
                reg_yml,
                comment=descr,
            )
            index += 1
            bitfields = reg.get_bitfields()
            if len(bitfields) > 0 and white_list and white_list[reg.name]:
                btf_yml = CM()
                for i, bitfield in enumerate(bitfields):
                    if diff and bitfield.get_value() == bitfield.get_reset_value():
                        continue
                    if (
                        white_list
                        and isinstance(white_list[reg.name], list)
                        and bitfield.name not in white_list[reg.name]
                    ):
                        continue
                    btf_yml.insert(
                        pos=i,
                        key=bitfield.name,
                        value=bitfield.get_enum_value(),
                        comment=self._get_bitfield_yaml_description(
                            bitfield, indent + SPSDK_YML_INDENT + SPSDK_YML_INDENT
                        ),
                    )
                reg_yml.insert(1, "bitfields", btf_yml, comment="The register bitfields")
            else:
                reg_yml.insert(
                    1,
                    "value",
                    reg.get_hex_value(),
                    comment=f"The value width: {reg.width}b",
                )

        return data

    # pylint: disable=(useless-param-doc, useless-type-doc)
    def create_yml_config(
        self,
        exclude_regs: Optional[List[str]] = None,
        exclude_fields: Optional[Dict[str, Dict[str, str]]] = None,
        ignored_fields: Optional[List[str]] = None,
        diff: bool = False,
        indent: int = 0,
    ) -> Any:
        """The function creates the configuration YML file.

        :param exclude_regs: List of excluded registers
        :param exclude_fields: Dictionary with lists of excluded bitfields per register
        :param ignored_fields: List of ignored names in fields.
        :param diff: Get only configuration with difference value to reset state.
        :param indent: Indent in space to generate YML.
        :return: YAML commented map of registers value.
        """
        white_list = {}
        for reg in self._registers:
            if exclude_regs and reg.name.startswith(tuple(exclude_regs)):
                continue
            value = []
            for bitfield in reg.get_bitfields():
                if ignored_fields and bitfield.name.startswith(tuple(ignored_fields)):
                    continue
                if (
                    exclude_fields
                    and reg.name in exclude_fields.keys()
                    and exclude_fields[reg.name]
                ):
                    if bitfield.name.startswith(tuple(exclude_fields[reg.name])):
                        continue
                value.append(bitfield.name)
            white_list[reg.name] = value if len(value) > 0 else None

        return self.create_yml_config_white_list(white_list if white_list else None, diff, indent)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2020-2022 NXP
#
# SPDX-License-Identifier: BSD-3-Clause

"""Miscellaneous functions in SBFile module."""

from datetime import datetime, timezone
from typing import Any, Sequence, Union

from spsdk import SPSDKError
from spsdk.utils import misc


class SecBootBlckSize:
    """Helper methods allowing to convert size to number of blocks and back.

    Note: The class is not intended to be instantiated
    """

    # Size of cipher block in bytes
    BLOCK_SIZE = 16

    @staticmethod
    def is_aligned(size: int) -> bool:
        """Whether size is aligned to cipher block size.

        :param size: given size in bytes
        :return: True if yes, False otherwise
        """
        return size % SecBootBlckSize.BLOCK_SIZE == 0

    @staticmethod
    def align(size: int) -> int:
        """Align given size to block size.

        :param size: in bytes
        :return: size aligned up to block size
        """
        return misc.align(size, SecBootBlckSize.BLOCK_SIZE)

    @staticmethod
    def to_num_blocks(size: int) -> int:
        """Converts size to number of cipher blocks.

        :param size: to be converted, the size must be aligned to block boundary
        :return: corresponding number of cipher blocks
        :raises SPSDKError: Raised when size is not aligned to block boundary
        """
        if not SecBootBlckSize.is_aligned(size):
            raise SPSDKError(
                f"Invalid size {size}, expected number aligned to BLOCK size {SecBootBlckSize.BLOCK_SIZE}"
            )
        return size // SecBootBlckSize.BLOCK_SIZE

    @staticmethod
    def align_block_fill_random(data: bytes) -> bytes:
        """Align block size to cipher block size.

        :param data: to be aligned
        :return: data aligned to cipher block size, filled with random values
        """
        return misc.align_block_fill_random(data, SecBootBlckSize.BLOCK_SIZE)


# the type represents input formats for BcdVersion3 value, see BcdVersion3.to_version
BcdVersion3Format = Union["BcdVersion3", str]


class BcdVersion3:
    """Version in format #.#.#, where # is BCD number (1-4 digits)."""

    # default value
    DEFAULT = "999.999.999"

    @staticmethod
    def _check_number(num: int) -> bool:
        """Check given number is a valid version number.

        :param num: to be checked
        :return: True if number format is valid
        :raises SPSDKError: If number format is not valid
        """
        if num < 0 or num > 0x9999:
            raise SPSDKError("Invalid number range")
        for index in range(4):
            if (num >> 4 * index) & 0xF > 0x9:
                raise SPSDKError("Invalid number, contains digit > 9")
        return True

    @staticmethod
    def _num_from_str(text: str) -> int:
        """Converts BCD number from text to int.

        :param text: given string to be converted to a version number
        :return: version number
        :raises SPSDKError: If format is not valid
        """
        if len(text) < 0 or len(text) > 4:
            raise SPSDKError("Invalid text length")
        result = int(text, 16)
        BcdVersion3._check_number(result)
        return result

    @staticmethod
    def from_str(text: str) -> "BcdVersion3":
        """Convert string to BcdVersion instance.

        :param text: version in format #.#.#, where # is 1-4 decimal digits
        :return: BcdVersion3 instance
        :raises SPSDKError: If format is not valid
        """
        parts = text.split(".")
        if len(parts) != 3:
            raise SPSDKError("Invalid length")
        major = BcdVersion3._num_from_str(parts[0])
        minor = BcdVersion3._num_from_str(parts[1])
        service = BcdVersion3._num_from_str(parts[2])
        return BcdVersion3(major, minor, service)

    @staticmethod
    def to_version(input_version: BcdVersion3Format) -> "BcdVersion3":
        """Convert different input formats into BcdVersion3 instance.

        :param input_version: either directly BcdVersion3 or string
        :raises SPSDKError: Raises when the format is unsupported
        :return: BcdVersion3 instance
        """
        if isinstance(input_version, BcdVersion3):
            return input_version
        if isinstance(input_version, str):
            return BcdVersion3.from_str(input_version)
        raise SPSDKError("unsupported format")

    def __init__(self, major: int = 1, minor: int = 0, service: int = 0):
        """Initialize BcdVersion3.

        :param major: number in BCD format, 1-4 decimal digits
        :param minor: number in BCD format, 1-4 decimal digits
        :param service: number in BCD format, 1-4 decimal digits
        :raises SPSDKError: Invalid version
        """
        if not all(
            [
                BcdVersion3._check_number(major),
                BcdVersion3._check_number(minor),
                BcdVersion3._check_number(service),
            ]
        ):
            raise SPSDKError("Invalid version")
        self.major = major
        self.minor = minor
        self.service = service

    def __str__(self) -> str:
        return f"{self.major:X}.{self.minor:X}.{self.service:X}"

    def __repr__(self) -> str:
        return self.__class__.__name__ + ": " + self.__str__()

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, BcdVersion3)
            and (self.major == other.major)
            and (self.minor == other.minor)
            and (self.service == other.service)
        )

    @property
    def nums(self) -> Sequence[int]:
        """Return array of version numbers: [major, minor, service]."""
        return [self.major, self.minor, self.service]


def pack_timestamp(value: datetime) -> int:
    """Converts datetime to millisecond since 1.1.2000.

    :param value: datetime to be converted
    :return: number of milliseconds since 1.1.2000  00:00:00; 64-bit integer
    :raises SPSDKError: When there is incorrect result of conversion
    """
    assert isinstance(value, datetime)
    start = datetime(2000, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc).timestamp()
    result = int((value.timestamp() - start) * 1000000)
    if result < 0 or result > 0xFFFFFFFFFFFFFFFF:
        raise SPSDKError("Incorrect result of conversion")
    return result


def unpack_timestamp(value: int) -> datetime:
    """Converts timestamp in milliseconds into datetime.

    :param value: number of milliseconds since 1.1.2000  00:00:00; 64-bit integer
    :return: corresponding datetime
    :raises SPSDKError: When there is incorrect result of conversion
    """
    assert isinstance(value, int)
    if value < 0 or value > 0xFFFFFFFFFFFFFFFF:
        raise SPSDKError("Incorrect result of conversion")
    start = int(datetime(2000, 1, 1, 0, 0, 0, 0, tzinfo=timezone.utc).timestamp() * 1000000)
    return datetime.fromtimestamp((start + value) / 1000000)



#!/usr/bin/env python
# -*- coding: UTF-8 -*-
#
# Copyright 2021-2023 NXP
#
# SPDX-License-Identifier: BSD-3-Clause
"""Utilities used in SmartCard."""

import logging
from typing import List, NamedTuple, Optional, Tuple

from spsdk import SPSDKError
from spsdk.apps.utils import format_raw_data
from spsdk.utils.easy_enum import Enum
from spsdk.utils.misc import get_hash

try:
    from smartcard.CardConnectionDecorator import CardConnection
    from smartcard.CardConnectionEvent import CardConnectionEvent
    from smartcard.CardConnectionObserver import CardConnectionObserver
    from smartcard.CardRequest import CardRequest
    from smartcard.CardType import ATRCardType
    from smartcard.System import readers
    from smartcard.util import toBytes
except ImportError as import_error:
    raise SPSDKError(
        "pyscard package is missing, please install it with pip install 'spsdk[tp]' in order to use TP"
    ) from import_error


from . import scard_commands

logger = logging.getLogger(__name__)


class ProvItem(Enum):
    """Provisioning item indexes."""

    OEM_CERT_COUNT = (0x1001, "oem_id_count")
    KEY_WRAP_ENABLE = (0x1002, "key_wrap_enable")
    ECC_DOMAIN_PARAM = (0x1003, "ecc_domain_param")
    JC_ID_PRK = (0x0000, "jc_id_prk")
    NXP_GLOB_PUK = (0x0001, "nxp_glob_puk")
    NXP_PROD_CERT = (0x0002, "nxp_prod_cert")
    OEM_ID_PRK = (0x0003, "oem_id_prk")
    OEM_LOG_PRK = (0x0004, "oem_log_prk")
    SB_KEK = (0x0005, "sb_kek")
    USER_KEK = (0x0006, "user_kek")
    CFPA = (0x0007, "cfpa")
    CMPA = (0x0008, "cmpa")
    PROD_COUNTER = (0x0009, "prod_counter")
    OEM_CERT_TEMPLATE = (0x0100, "oem_cert_template")


class LoggerConnectionObserver(CardConnectionObserver):
    """Overriding standard CardConnectionObserver, to enable proper logging."""

    def update(
        self, cardconnection: CardConnection, cardconnectionevent: CardConnectionEvent
    ) -> None:
        """New card connection event. Just log it.

        :param cardconnection: Card connection object.
        :param cardconnectionevent: Card connection event.
        """
        reader = cardconnection.getReader()
        if cardconnectionevent.type == "connect":
            logger.debug(f"connecting to {reader}")
        elif cardconnectionevent.type == "disconnect":
            logger.debug(f"disconnecting from {reader}")
        elif cardconnectionevent.type == "command":
            command_data = bytes(cardconnectionevent.args[0])
            if len(command_data) < 16:
                logger.debug(f"> {format_raw_data(command_data)}")
            else:
                data_start = 7 if command_data[4] == 0 else 5
                logger.debug(
                    f"> {format_raw_data(command_data[:data_start])}\n"
                    f"{format_raw_data(command_data[data_start:], use_hexdump=True)}"
                )
        elif cardconnectionevent.type == "response":
            if [] == cardconnectionevent.args[0]:
                logger.debug(
                    f"< {format(cardconnectionevent.args[1], '02x')} {format(cardconnectionevent.args[2], '02x')}"
                )
            else:
                logger.debug(
                    f"< data:\n{format_raw_data(bytes(cardconnectionevent.args[0]), use_hexdump=True)}",
                )
                logger.debug(
                    f"< {format(cardconnectionevent.args[1], '02x')} {format(cardconnectionevent.args[2], '02x')}",
                )
        else:
            logger.info(f"Something Interesting{cardconnectionevent}")


class AppletInfo(NamedTuple):
    """Simple storage for information about a reader/card/applet."""

    reader_name: str
    version: str
    serial_number: int
    card_connection: CardConnection


def get_readers() -> List[Tuple[str, str]]:
    """Return list of all readers in the system.

    Each reader is represented by a tuple:
        1. item: card reader name
        2. item: hash of the card reader's name
    """
    return [(str(reader), get_hash(str(reader))) for reader in readers()]


def get_applet_infos(
    atr: str, applet: str, filter_id: Optional[int] = None, filter_reader: Optional[str] = None
) -> List[AppletInfo]:
    """Collets information about attached card readers.

    :param atr: Select the card's ATR (after reset value)
    :param applet: Name of an applet to look for
    :param filter_id: Filter the card's ID (serial number), defaults to None
    :param filter_reader: Name of the preferred reader
    :return: List of card/applets fulfilling the search criteria.
    """
    ret = []
    card_readers = get_readers()
    for reader_tuple in card_readers:
        if filter_reader and filter_reader not in reader_tuple:
            continue
        logger.info(f"Checking reader {reader_tuple}")
        try:
            request = CardRequest(
                readers=[reader_tuple[0]],
                cardType=ATRCardType(toBytes(atr)),
                timeout=0.1,
            )
            scard = request.waitforcard()
        # pylint: disable=broad-except
        except Exception as e:
            # if anything goes wrong, just log the error and move on to next reader
            logger.debug(str(e))
            continue
        assert isinstance(scard.connection, CardConnection)
        try:
            scard.connection.addObserver(LoggerConnectionObserver())
            scard.connection.connect()
            select_cmd = scard_commands.Select(applet=applet)
            select_cmd.transmit(scard.connection)
            serial_number_cmd = scard_commands.GetSerialNumber()
            serial_number = serial_number_cmd.format(
                serial_number_cmd.transmit(scard.connection),
            )
            if filter_id and str(filter_id) != str(serial_number):
                continue
            version_cmd = scard_commands.GetAppletVersion()
            version = version_cmd.format(version_cmd.transmit(scard.connection))
            ret.append(
                AppletInfo(
                    reader_name=reader_tuple[0],
                    version=version,
                    serial_number=serial_number,
                    card_connection=scard.connection,
                )
            )
        # pylint: disable=broad-except
        except Exception as e:
            # if anything goes wrong, just log the error and move on to next reader
            logger.debug(str(e))
            scard.connection.disconnect()
            continue

    return ret
